@@ -3,7 +3,7 @@
 /* header file for S-Lang internal structures that users do not (should not)
    need.  Use slang.h for that purpose. */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004-2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -91,8 +91,8 @@ typedef enum
    SLANG_BC_LVARIABLE_APUT	= 0x29,
    SLANG_BC_LOBJPTR		= 0x2A,
    SLANG_BC_GOBJPTR		= 0x2B,
-   SLANG_BC_UNUSED_0x2C		= 0x2C,
-   SLANG_BC_UNUSED_0x2D		= 0x2D,
+   SLANG_BC_FIELD_REF		= 0x2C,
+   SLANG_BC_DEREF_FUN_CALL	= 0x2D,
    SLANG_BC_UNUSED_0x2E		= 0x2E,
    SLANG_BC_UNUSED_0x2F		= 0x2F,
    SLANG_BC_UNUSED_0x30		= 0x30,
@@ -346,11 +346,17 @@ _pSLang_BC_Type;
 
 #define SLANG_BCST_IF		0x20
 #define SLANG_BCST_IFNOT	0x21
-#define SLANG_BCST_ELSE	0x22
+#define SLANG_BCST_ELSE		0x22
 #define SLANG_BCST_ANDELSE	0x23
 #define SLANG_BCST_ORELSE	0x24
 #define SLANG_BCST_SWITCH	0x25
 #define SLANG_BCST_NOTELSE	0x26
+#define SLANG_BCST_SC_OR	0x27
+#define SLANG_BCST_SC_AND	0x28
+
+#define SLANG_BCST_LOOP_ELSE	0x29
+#define SLANG_BCST_LOOP_THEN	0x30
+#define SLANG_BCST_COMPARE      0x31
 
 /* assignment (SLANG_BC_SET_*_LVALUE) subtypes.  The order MUST correspond
  * to the assignment token order with the ASSIGN_TOKEN as the first!
@@ -540,28 +546,35 @@ typedef struct _pSLang_Struct_Type
 }
 _pSLang_Struct_Type;
 
-extern void _pSLstruct_delete_struct (_pSLang_Struct_Type *);
-extern int _pSLang_push_struct (_pSLang_Struct_Type *);
-extern int _pSLang_pop_struct (_pSLang_Struct_Type **);
 extern int _pSLstruct_init (void);
-/* extern int _pSLstruct_get_field (char *); */
 extern int _pSLstruct_define_struct (void);
+extern int _pSLstruct_define_struct2 (void);
 extern int _pSLstruct_define_typedef (void);
 
+extern SLang_Object_Type *_pSLstruct_get_field_value (SLang_Struct_Type *, char *);
+extern int _pSLstruct_push_field_ref (char *);
+
+int _pSLang_get_qualifiers (SLang_Struct_Type **);
+
 struct _pSLang_Ref_Type
 {
-   int is_global;
-   union
-     {
-	SLang_Name_Type *nt;
-	SLang_Object_Type *local_obj;
-     }
-   v;
+   int num_refs;
+   VOID_STAR data;
+   unsigned int sizeof_data;
+   int data_is_nametype;	       /* used for optimization */
+   int (*deref_assign)(VOID_STAR);
+   int (*deref) (VOID_STAR);
+   char *(*string)(VOID_STAR);		       /* returns a malloced string */
+   void (*destroy) (VOID_STAR);
+   int (*is_initialized) (VOID_STAR);
+   int (*uninitialize) (VOID_STAR);
 };
 
 extern int _pSLang_dereference_ref (SLang_Ref_Type *);
 extern int _pSLang_deref_assign (SLang_Ref_Type *);
-extern int _pSLang_push_ref (int, VOID_STAR);
+extern int SLang_push_ref (SLang_Ref_Type *);
+extern int _pSLang_push_nt_as_ref (SLang_Name_Type *);
+extern SLang_Ref_Type *_pSLang_new_ref (unsigned int);
 
 extern int _pSL_increment_frame_pointer (void);
 extern int _pSL_decrement_frame_pointer (void);
@@ -631,6 +644,8 @@ extern int _pSLpush_slang_obj (SLang_Object_Type *);
 extern char *_pSLexpand_escaped_char(char *, SLwchar_Type *, int *);
 extern void _pSLexpand_escaped_string (char *, char *, char *);
 
+extern int _pSLpush_dollar_string (char *);
+
 /* returns a pointer to an SLstring string-- use SLang_free_slstring */
 extern char *_pSLstringize_object (SLang_Object_Type *);
 extern int _pSLdump_objects (char *, SLang_Object_Type *, unsigned int, int);
@@ -728,6 +743,7 @@ extern unsigned int _pSLpc_convert_scancode (unsigned int, unsigned int, int);
 
 typedef struct _pSLterminfo_Type SLterminfo_Type;
 extern SLterminfo_Type *_pSLtt_tigetent (char *);
+extern void _pSLtt_tifreeent (SLterminfo_Type *);
 extern char *_pSLtt_tigetstr (SLterminfo_Type *, char *);
 extern int _pSLtt_tigetnum (SLterminfo_Type *, char *);
 extern int _pSLtt_tigetflag (SLterminfo_Type *, char *);
@@ -919,6 +935,7 @@ extern void _pSLang_set_class_info (SLtype, SLang_Class_Type *);
 extern int _pSLarith_bin_op (SLang_Object_Type *, SLang_Object_Type *, int);
 
 extern int _pSLarray_add_bin_op (SLtype);
+extern int _pSLarray_push_elem_ref (void);
 
 typedef struct
 {
@@ -942,6 +959,8 @@ typedef struct
 SLang_Arith_Binary_Type;
 extern int _pSLadd_arith_binary_table (SLang_Arith_Binary_Type *tbl, char *);
 
+extern int _pSLang_do_binary_ab (int op, SLang_Object_Type *obja, SLang_Object_Type *objb);
+
 extern int _pSLang_call_funptr (SLang_Name_Type *);
 extern void _pSLset_double_format (char *);
 extern SLang_Name_Type *_pSLlocate_global_name (char *);
@@ -966,6 +985,7 @@ extern int _pSLarray_aput (void);
 extern int _pSLarray_aget (void);
 extern int _pSLarray_aget1 (unsigned int);
 extern int _pSLarray_inline_implicit_array (void);
+extern int _pSLarray_inline_implicit_arrayn (void);
 extern int _pSLarray_inline_array (void);
 extern int _pSLarray_wildcard_array (void);
 
@@ -995,6 +1015,7 @@ SLang_create_array1 (SLtype, int, VOID_STAR, SLindex_Type *, unsigned int, int);
 
 extern int _pSLassoc_aput (SLtype, unsigned int);
 extern int _pSLassoc_aget (SLtype, unsigned int);
+extern int _pSLassoc_inc_value (unsigned int, int val);
 
 extern int _pSLcompile_push_context (SLang_Load_Type *);
 extern int _pSLcompile_pop_context (void);
@@ -1039,7 +1060,7 @@ extern void (*_pSLcompile_ptr)(_pSLang_Token_Type *);
 extern char *_pSLskip_whitespace (char *s);
 
 /* slospath.c */
-extern char *_pSLpath_find_file (char *);   /* slstring returned */
+extern char *_pSLpath_find_file (char *, int);   /* slstring returned */
 
 /* Read but do not set this variable. */
 extern volatile int _pSLang_Error;
@@ -1071,6 +1092,7 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define NOP_TOKEN	0x05
 #define FARG_TOKEN	0x06
 #define TMP_TOKEN	0x07
+#define QUALIFIER_TOKEN	0x08
 
 #define RESERVED1_TOKEN	0x0A	       /* \n */
 #define RESERVED2_TOKEN	0x0D	       /* \r */
@@ -1124,20 +1146,24 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define NAMESPACE_TOKEN	0x34
 
 /* Operators */
-#define POW_TOKEN	0x38
-
 /* The order here must match the order in the Binop_Level table in slparse.c */
-#define FIRST_BINARY_OP	0x39
+#define FIRST_BINARY_OP	0x36
+#define SC_AND_TOKEN	0x36
+#define SC_OR_TOKEN	0x37
+#define POW_TOKEN	0x38
 #define ADD_TOKEN	0x39
 #define SUB_TOKEN	0x3a
 #define TIMES_TOKEN	0x3b
 #define DIV_TOKEN	0x3c
+
 #define LT_TOKEN	0x3d
 #define LE_TOKEN	0x3e
 #define GT_TOKEN	0x3f
 #define GE_TOKEN	0x40
 #define EQ_TOKEN	0x41
 #define NE_TOKEN	0x42
+#define IS_COMPARE_OP(x) (((x) >= LT_TOKEN) && ((x) <= NE_TOKEN))
+
 #define AND_TOKEN	0x43
 #define OR_TOKEN	0x44
 #define MOD_TOKEN	0x45
@@ -1147,7 +1173,6 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define BXOR_TOKEN	0x49
 #define BOR_TOKEN	0x4a
 #define POUND_TOKEN	0x4b	       /* matrix multiplication */
-
 #define LAST_BINARY_OP	 0x4b
 #define IS_BINARY_OP(t)	 ((t >= FIRST_BINARY_OP) && (t <= LAST_BINARY_OP))
 
@@ -1184,7 +1209,6 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define POST_MINUSMINUS_TOKEN	0x61
 
 /* Directives */
-#define FIRST_DIRECTIVE_TOKEN	0x62
 #define IFNOT_TOKEN	0x62
 #define IF_TOKEN	0x63
 #define ELSE_TOKEN	0x64
@@ -1231,6 +1255,10 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define THROW_TOKEN	0x89
 #define FINALLY_TOKEN	0x8a
 
+#define THEN_TOKEN	0x8d
+
+#define STRUCT_WITH_ASSIGN_TOKEN	0x8e
+
 /* Note: the order here must match the order of the generic assignment tokens.
  * Also, the first token of each group must be the ?_ASSIGN_TOKEN.
  * slparse.c exploits this order, as well as slang.h.
@@ -1287,11 +1315,21 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define LAST_ASSIGN_TOKEN		0xCA
 #define IS_ASSIGN_TOKEN(t) (((t)>=FIRST_ASSIGN_TOKEN)&&((t)<=LAST_ASSIGN_TOKEN))
 
+#define _DEREF_FUNCALL_TOKEN		0xCE
+
+#define LOOP_THEN_TOKEN			0xD0
+/* #define LOOP_ELSE_TOKEN			0xD1 */
+
+#define _COMPARE_TOKEN			0xD8   /* (a < b < c ...) */
+#define _ARRAY_ELEM_REF_TOKEN		0xD9   /* &X[i] */
+#define _STRUCT_FIELD_REF_TOKEN		0xDA   /* &S.a */
+
 #define _INLINE_ARRAY_TOKEN		0xE0
-#define _INLINE_IMPLICIT_ARRAY_TOKEN	0xE1
+#define _INLINE_IMPLICIT_ARRAY_TOKEN	0xE1   /* [a:b:c] */
 #define _NULL_TOKEN			0xE2
 #define _INLINE_WILDCARD_ARRAY_TOKEN	0xE3
 #define _INLINE_LIST_TOKEN		0xE4
+#define _INLINE_IMPLICIT_ARRAYN_TOKEN	0xE5   /* [a:b:#n] */
 
 #define ESC_STRING_DOLLAR_TOKEN		0xF0
 
@@ -1345,18 +1383,25 @@ extern int _pSLerr_suspend_messages (void);
 extern int _pSLerr_resume_messages (void);
 extern int _pSLerr_traceback_msg (char *, ...) SLATTRIBUTE_PRINTF(1,2);
 extern void _pSLerr_dump_msg (char *, ...) SLATTRIBUTE_PRINTF(1,2);
-extern void _pSLerr_clear_error (void);
+extern void _pSLerr_clear_error (int);
+
+typedef struct _pSLerr_Error_Queue_Type _pSLerr_Error_Queue_Type;
+extern _pSLerr_Error_Queue_Type *_pSLerr_new_error_queue (int);
+extern int _pSLerr_set_error_queue (_pSLerr_Error_Queue_Type *);
+extern void _pSLerr_delete_error_queue (_pSLerr_Error_Queue_Type *);
+
 extern void _pSLerr_print_message_queue (void);
 
 extern int _pSLerr_set_line_info (char *, int, char *);
-extern int _pSLang_pop_error_context (void);
+extern int _pSLang_pop_error_context (int);
 extern int _pSLang_push_error_context (void);
 extern void (*_pSLinterpreter_Error_Hook)(int);
 extern int _pSLerr_init_interp_exceptions (void);
 extern int _pSLerr_get_last_error (void);
 extern int _pSLerr_pop_exception (int *);
 extern void _pSLerr_free_queued_messages (void);
-extern char *_pSLerr_get_error_from_queue (void);
+
+extern char *_pSLerr_get_error_from_queue (_pSLerr_Error_Queue_Type *);
 extern int _pSLerr_throw (void);
 extern int (*_pSLerr_New_Exception_Hook)(char *name, char *desc, int error_code);
 
@@ -1381,6 +1426,9 @@ extern int _pSLang_get_frame_depth (void);
 
 #if SLANG_HAS_FLOAT
 extern int _pSLmath_isnan (double x);
+extern int _pSLmath_isinf (double x);
+extern double _pSLang_NaN;
+extern double _pSLang_Inf;
 #endif
 
 #if SLANG_HAS_SIGNALS
