@@ -2,7 +2,7 @@
 #define DAVIS_SLANG_H_
 /* -*- mode: C; mode: fold; -*- */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004-2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -22,8 +22,8 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 USA.  
 */
 
-#define SLANG_VERSION 20007
-#define SLANG_VERSION_STRING "2.0.7"
+#define SLANG_VERSION 20103
+#define SLANG_VERSION_STRING "2.1.3"
 /* #ifdef __DATE__ */
 /* # define SLANG_VERSION_STRING SLANG_VERSION_STRING0 " " __DATE__ */
 /* #else */
@@ -49,6 +49,12 @@ USA.
 # endif
 #endif
 
+#if defined(__APPLE__)
+# ifndef __unix__
+#  define __unix__ 1		       /* cough */
+# endif
+#endif
+  
 #if !defined(__GO32__)
 # ifdef __unix__
 #  define REAL_UNIX_SYSTEM
@@ -231,7 +237,7 @@ typedef int (*FVOID_STAR)(void);
 # endif
 #endif
 
-#if defined(__WIN32__) && SLANG_DLL
+#if (defined(__WIN32__) || defined(__CYGWIN32__)) && defined(SLANG_DLL) && SLANG_DLL
 # define SL_EXPORT __declspec(dllexport)
 # define SL_IMPORT __declspec(dllimport)
 #else
@@ -325,13 +331,11 @@ SL_EXTERN int SLwchar_isxdigit (SLwchar_Type);
 #define SLWCWIDTH_SINGLE_WIDTH	0x0001
 SL_EXTERN int SLwchar_set_wcwidth_flags (int flags);
 
+/* The skip_char functions also skip over malformed strings counting each byte
+ * in the malformed sequence.
+ */
 SL_EXTERN SLuchar_Type *SLutf8_skip_char (SLuchar_Type *u, SLuchar_Type *umax);
 SL_EXTERN SLuchar_Type *SLutf8_bskip_char (SLuchar_Type *umin, SLuchar_Type *u);
-
-/* The SLutf8_strup/lo functions return slstrings -- free with SLang_free_slstring */
-SL_EXTERN SLuchar_Type *SLutf8_strup (SLuchar_Type *u, SLuchar_Type *umax);
-SL_EXTERN SLuchar_Type *SLutf8_strlo (SLuchar_Type *u, SLuchar_Type *umax);
-
 SL_EXTERN SLuchar_Type *SLutf8_skip_chars (SLuchar_Type *u, SLuchar_Type *umax,
 					unsigned int num, unsigned int *dnum,
 					int ignore_combining );
@@ -339,6 +343,10 @@ SL_EXTERN SLuchar_Type *SLutf8_bskip_chars (SLuchar_Type *umin, SLuchar_Type *u,
 					  unsigned int num, unsigned int *dnum,
 					 int ignore_combining);
 
+/* The SLutf8_strup/lo functions return slstrings -- free with SLang_free_slstring */
+SL_EXTERN SLuchar_Type *SLutf8_strup (SLuchar_Type *u, SLuchar_Type *umax);
+SL_EXTERN SLuchar_Type *SLutf8_strlo (SLuchar_Type *u, SLuchar_Type *umax);
+
 SL_EXTERN SLstr_Type *SLutf8_subst_wchar (SLuchar_Type *u, SLuchar_Type *umax,
 				       SLwchar_Type wch, unsigned int pos,
 				       int ignore_combining);
@@ -594,6 +602,7 @@ SL_EXTERN void SLns_delete_namespace (SLang_NameSpace_Type *);
 SL_EXTERN int SLns_load_file (char *, char *);
 SL_EXTERN int SLns_load_string (char *, char *);
 SL_EXTERN int (*SLns_Load_File_Hook) (char *, char *);
+
 SL_EXTERN int SLang_load_file_verbose (int);    
 /* if non-zero, display file loading messages */
 
@@ -836,6 +845,7 @@ SL_EXTERN void SLang_inc_mmt (SLang_MMT_Type *);
 typedef int SLindex_Type;
 typedef unsigned int SLuindex_Type;
 #define SLANG_ARRAY_INDEX_TYPE SLANG_INT_TYPE
+SL_EXTERN int SLang_pop_array_index (SLindex_Type *);
 typedef struct _pSLang_Array_Type
 {
    SLtype data_type;
@@ -861,10 +871,10 @@ typedef struct _pSLang_Array_Type
 }
 SLang_Array_Type;
 
-SL_EXTERN int SLang_pop_array_of_type (SLang_Array_Type **, SLtype);
-SL_EXTERN int SLang_pop_array (SLang_Array_Type **, int);
-SL_EXTERN int SLang_push_array (SLang_Array_Type *, int);
-SL_EXTERN void SLang_free_array (SLang_Array_Type *);
+SL_EXTERN int SLang_pop_array_of_type (SLang_Array_Type **atp, SLtype type);
+SL_EXTERN int SLang_pop_array (SLang_Array_Type **atp, int convert_scalar);
+SL_EXTERN int SLang_push_array (SLang_Array_Type *at, int do_free);
+SL_EXTERN void SLang_free_array (SLang_Array_Type *at);
 SL_EXTERN SLang_Array_Type *SLang_create_array (SLtype, int, VOID_STAR, SLindex_Type *, unsigned int);
 SL_EXTERN SLang_Array_Type *SLang_duplicate_array (SLang_Array_Type *);
 SL_EXTERN int SLang_get_array_element (SLang_Array_Type *, SLindex_Type *, VOID_STAR);
@@ -903,8 +913,9 @@ SL_EXTERN int SLarray_map_array (SLCONST SLarray_Map_Type *);
 
 /*{{{ Interpreter Function Prototypes */
 
-/* SL_EXTERN int SLerr_throw (int err, char *msg, SLtype obj_type, VOID_STAR objptr); */
+SL_EXTERN int SLerr_throw (int err, char *msg, SLtype obj_type, VOID_STAR objptr);
 SL_EXTERN void SLang_verror (int, char *, ...) SLATTRIBUTE_PRINTF(2,3);
+SL_EXTERN void SLang_verror_va (int errcode, char *fmt, va_list va);
 SL_EXTERN int SLang_get_error (void);
 SL_EXTERN int SLang_set_error (int);
 SL_EXTERN char *SLerr_strerror (int errcode);
@@ -990,6 +1001,11 @@ SL_EXTERN int SL_Import_Error;
 #define SL_INVALID_UTF8			SL_InvalidUTF8_Error
 
   SL_EXTERN int SLang_Traceback;
+#define SL_TB_NONE		0x0
+#define SL_TB_FULL		0x1    /* full traceback */
+#define SL_TB_OMIT_LOCALS	0x2    /* full, but omit local vars */
+#define SL_TB_PARTIAL		0x4    /* show just on line of traceback */
+
   /* If non-zero, dump an S-Lang traceback upon error.  Available as
      _traceback in S-Lang. */
 
@@ -1495,6 +1511,7 @@ SL_EXTERN int SLang_Key_TimeOut_Flag;
 
 typedef struct _pSLrline_Type SLrline_Type;
 SL_EXTERN SLrline_Type *SLrline_open (unsigned int width, unsigned int flags);
+SL_EXTERN SLrline_Type *SLrline_open2 (char *, unsigned int width, unsigned int flags);
 #define SL_RLINE_NO_ECHO	1
 #define SL_RLINE_USE_ANSI	2
 #define SL_RLINE_BLINK_MATCH	4
@@ -1508,6 +1525,7 @@ SL_EXTERN int SLrline_bol (SLrline_Type *);
 SL_EXTERN int SLrline_eol (SLrline_Type *);
 SL_EXTERN int SLrline_del (SLrline_Type *, unsigned int len);
 SL_EXTERN int SLrline_ins (SLrline_Type *, char *s, unsigned int len);
+SL_EXTERN int SLrline_move (SLrline_Type *, int);
 
 SL_EXTERN int SLrline_set_echo (SLrline_Type *, int);
 SL_EXTERN int SLrline_set_tab (SLrline_Type *, unsigned int tabwidth);
@@ -1536,6 +1554,9 @@ SL_EXTERN void SLrline_redraw (SLrline_Type *);
 SL_EXTERN int SLrline_save_line (SLrline_Type *);
 SL_EXTERN int SLrline_add_to_history (SLrline_Type *, char *);
 
+/* Interpreter interface */
+SL_EXTERN int SLrline_init (char *appdef, char *user_initfile, char *sys_initfile);
+
 /* Compatibility */
 typedef SLrline_Type SLang_RLine_Info_Type;
 
@@ -2120,6 +2141,16 @@ SL_EXTERN int SLadd_intrinsic_function (char *, FVOID_STAR, SLtype, unsigned int
 SL_EXTERN int SLns_add_intrinsic_variable (SLang_NameSpace_Type *, char *, VOID_STAR, SLtype, int);
 SL_EXTERN int SLns_add_intrinsic_function (SLang_NameSpace_Type *, char *, FVOID_STAR, SLtype, unsigned int,...);
 
+/* These functions are used to patch intrinsic tables that make use of 
+ * dynamically allocated types whose type id is determined at
+ * run-time.  The second version is useful for the most common case of a 
+ * single mapping.
+ */
+SL_EXTERN int SLclass_patch_intrin_fun_table (SLang_Intrin_Fun_Type *table, 
+					 SLtype *from_types, SLtype *to_types, unsigned int num);
+SL_EXTERN int SLclass_patch_intrin_fun_table1 (SLang_Intrin_Fun_Type *table, 
+					  SLtype from_type, SLtype to_type);
+
 #define MAKE_INTRINSIC_N(n,f,out,in,a1,a2,a3,a4,a5,a6,a7) \
     {(n), NULL, SLANG_INTRINSIC, (FVOID_STAR) (f), \
       {a1,a2,a3,a4,a5,a6,a7}, (in), (out)}
@@ -2281,7 +2312,7 @@ SL_EXTERN int SLregexp_get_hints (SLRegexp_Type *, unsigned int *flagsp);
 #define SLREGEXP_HINT_BOL		0x01   /* pattern must match bol */
 #define SLREGEXP_HINT_OSEARCH		0x02   /* ordinary search will do */
 
-SL_EXTERN char *SLregexp_quote_string (char *, char *, unsigned int);
+SL_EXTERN char *SLregexp_quote_string (char *pattern, char *buf, unsigned int buflen);
 
 /*}}}*/
 
