@@ -1,10 +1,24 @@
 /* Complex Data Type definition for S-Lang */
-/* Copyright (c) 1997, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -30,7 +44,7 @@ int SLang_pop_complex (double *r, double *i)
    switch (SLang_peek_at_stack ())
      {
       case SLANG_COMPLEX_TYPE:
-	if (-1 == SLclass_pop_ptr_obj (SLANG_COMPLEX_TYPE, (VOID_STAR *)&c))
+	if (-1 == SLclass_pop_ptr_obj (SLANG_COMPLEX_TYPE, VOID_STAR_STAR(&c)))
 	  return -1;
 	*r = c[0];
 	*i = c[1];
@@ -39,7 +53,7 @@ int SLang_pop_complex (double *r, double *i)
 
       default:
 	*i = 0.0;
-	if (-1 == SLang_pop_double (r, NULL, NULL))
+	if (-1 == SLang_pop_double (r))
 	  return -1;
 	break;
 
@@ -114,11 +128,23 @@ double *SLcomplex_divide (double *c, double *a, double *b)
 /* a^b = exp (b log a); */
 double *SLcomplex_pow (double *c, double *a, double *b)
 {
+   if ((a[0] == 0.0) && (b[0] == 0.0) && (a[1] == 0.0) && (b[1] == 0.0))
+     {
+	c[0] = 1.0;
+	c[1] = 0.0;
+	return c;
+     }
    return SLcomplex_exp (c, SLcomplex_times (c, b, SLcomplex_log (c, a)));
 }
 
 static double *complex_dpow (double *c, double *a, double b)
 {
+   if ((b == 0.0) && (a[0] == 0.0) && (a[1] == 0.0))
+     {
+	c[0] = 1.0;
+	c[1] = 0.0;
+	return c;
+     }
    SLcomplex_log (c, a);
    c[0] *= b;
    c[1] *= b;
@@ -127,6 +153,12 @@ static double *complex_dpow (double *c, double *a, double b)
 
 static double *dcomplex_pow (double *c, double a, double *b)
 {
+   if ((a == 0.0) && (b[0] == 0.0) && (b[1] == 0.0))
+     {
+	c[0] = 1.0;
+	c[1] = 0.0;
+	return c;
+     }
    a = log (a);
    c[0] = a * b[0];
    c[1] = a * b[1];
@@ -407,8 +439,8 @@ double *SLcomplex_atanh (double *atanhz, double *z)
    return atanhz;
 }
 
-static int complex_binary_result (int op, unsigned char a, unsigned char b,
-				  unsigned char *c)
+static int complex_binary_result (int op, SLtype a, SLtype b,
+				  SLtype *c)
 {
    (void) a; (void) b;
 
@@ -432,8 +464,8 @@ static int complex_binary_result (int op, unsigned char a, unsigned char b,
 }
 
 static int complex_complex_binary (int op,
-				   unsigned char a_type, VOID_STAR ap, unsigned int na,
-				   unsigned char b_type, VOID_STAR bp, unsigned int nb,
+				   SLtype a_type, VOID_STAR ap, unsigned int na,
+				   SLtype b_type, VOID_STAR bp, unsigned int nb,
 				   VOID_STAR cp)
 {
    char *ic;
@@ -489,11 +521,13 @@ static int complex_complex_binary (int op,
       case SLANG_DIVIDE:	       /* / */
 	for (n = 0; n < n_max; n += 2)
 	  {
+#if 0
 	     if ((b[0] == 0.0) && (b[1] == 0.0))
 	       {
-		  SLang_Error = SL_DIVIDE_ERROR;
+		  SLang_set_error (SL_DIVIDE_ERROR);
 		  return -1;
 	       }
+#endif
 	     SLcomplex_divide (c + n, a, b);
 	     a += da; b += db;
 	  }
@@ -529,8 +563,8 @@ static int complex_complex_binary (int op,
 }
 
 static int complex_double_binary (int op,
-				  unsigned char a_type, VOID_STAR ap, unsigned int na,
-				  unsigned char b_type, VOID_STAR bp, unsigned int nb,
+				  SLtype a_type, VOID_STAR ap, unsigned int na,
+				  SLtype b_type, VOID_STAR bp, unsigned int nb,
 				  VOID_STAR cp)
 {
    char *ic;
@@ -589,11 +623,13 @@ static int complex_double_binary (int op,
 	for (n = 0; n < n_max; n += 2)
 	  {
 	     double b0 = b[0];
+#if 0
 	     if (b0 == 0.0)
 	       {
-		  SLang_Error = SL_DIVIDE_ERROR;
+		  SLang_set_error (SL_DIVIDE_ERROR);
 		  return -1;
 	       }
+#endif
 	     c[n] = a[0] / b0;
 	     c[n + 1] = a[1] / b0;
 	     a += da; b += db;
@@ -629,8 +665,8 @@ static int complex_double_binary (int op,
 }
 
 static int double_complex_binary (int op,
-				  unsigned char a_type, VOID_STAR ap, unsigned int na,
-				  unsigned char b_type, VOID_STAR bp, unsigned int nb,
+				  SLtype a_type, VOID_STAR ap, unsigned int na,
+				  SLtype b_type, VOID_STAR bp, unsigned int nb,
 				  VOID_STAR cp)
 {
    char *ic;
@@ -689,11 +725,13 @@ static int double_complex_binary (int op,
 	for (n = 0; n < n_max; n += 2)
 	  {
 	     double z[2];
+#if 0
 	     if ((b[0] == 0.0) && (b[1] == 0.0))
 	       {
-		  SLang_Error = SL_DIVIDE_ERROR;
+		  SLang_set_error (SL_DIVIDE_ERROR);
 		  return -1;
 	       }
+#endif
 	     z[0] = a[0];
 	     z[1] = 0.0;
 	     SLcomplex_divide (c + n, z, b);
@@ -730,8 +768,8 @@ static int double_complex_binary (int op,
 }
 
 static int complex_generic_binary (int op,
-				   unsigned char a_type, VOID_STAR ap, unsigned int na,
-				   unsigned char b_type, VOID_STAR bp, unsigned int nb,
+				   SLtype a_type, VOID_STAR ap, unsigned int na,
+				   SLtype b_type, VOID_STAR bp, unsigned int nb,
 				   VOID_STAR cp)
 {
    char *ic;
@@ -803,11 +841,13 @@ static int complex_generic_binary (int op,
 	for (n = 0; n < n_max; n += 2)
 	  {
 	     double b0 = to_double((VOID_STAR)b);
+#if 0
 	     if (b0 == 0)
 	       {
-		  SLang_Error = SL_DIVIDE_ERROR;
+		  SLang_set_error (SL_DIVIDE_ERROR);
 		  return -1;
 	       }
+#endif
 	     c[n] = a[0] / b0;
 	     c[n + 1] = a[1] / b0;
 	     a += da; b += db;
@@ -835,8 +875,8 @@ static int complex_generic_binary (int op,
 }
 
 static int generic_complex_binary (int op,
-				   unsigned char a_type, VOID_STAR ap, unsigned int na,
-				   unsigned char b_type, VOID_STAR bp, unsigned int nb,
+				   SLtype a_type, VOID_STAR ap, unsigned int na,
+				   SLtype b_type, VOID_STAR bp, unsigned int nb,
 				   VOID_STAR cp)
 {
    double *b, *c;
@@ -906,11 +946,13 @@ static int generic_complex_binary (int op,
 	for (n = 0; n < n_max; n += 2)
 	  {
 	     double z[2];
+#if 0
 	     if ((b[0] == 0.0) && (b[1] == 0.0))
 	       {
-		  SLang_Error = SL_DIVIDE_ERROR;
+		  SLang_set_error (SL_DIVIDE_ERROR);
 		  return -1;
 	       }
+#endif
 	     z[0] = to_double((VOID_STAR)a);
 	     z[1] = 0.0;
 	     SLcomplex_divide (c + n, z, b);
@@ -938,7 +980,7 @@ static int generic_complex_binary (int op,
    return 1;
 }
 
-static int complex_unary_result (int op, unsigned char a, unsigned char *b)
+static int complex_unary_result (int op, SLtype a, SLtype *b)
 {
    (void) a;
 
@@ -967,7 +1009,7 @@ static int complex_unary_result (int op, unsigned char a, unsigned char *b)
 }
 
 static int complex_unary (int op,
-			  unsigned char a_type, VOID_STAR ap, unsigned int na,
+			  SLtype a_type, VOID_STAR ap, unsigned int na,
 			  VOID_STAR bp)
 {
    unsigned int n;
@@ -1035,8 +1077,8 @@ static int complex_unary (int op,
 }
 
 static int
-complex_typecast (unsigned char from_type, VOID_STAR from, unsigned int num,
-		  unsigned char to_type, VOID_STAR to)
+complex_typecast (SLtype from_type, VOID_STAR from, unsigned int num,
+		  SLtype to_type, VOID_STAR to)
 {
    double *z;
    double *d;
@@ -1077,13 +1119,13 @@ complex_typecast (unsigned char from_type, VOID_STAR from, unsigned int num,
    return 1;
 }
 
-static void complex_destroy (unsigned char type, VOID_STAR ptr)
+static void complex_destroy (SLtype type, VOID_STAR ptr)
 {
    (void) type;
    SLfree ((char *)*(double **) ptr);
 }
 
-static int complex_push (unsigned char type, VOID_STAR ptr)
+static int complex_push (SLtype type, VOID_STAR ptr)
 {
    double *z;
 
@@ -1092,7 +1134,7 @@ static int complex_push (unsigned char type, VOID_STAR ptr)
    return SLang_push_complex (z[0], z[1]);
 }
 
-static int complex_pop (unsigned char type, VOID_STAR ptr)
+static int complex_pop (SLtype type, VOID_STAR ptr)
 {
    double *z;
 
@@ -1101,10 +1143,10 @@ static int complex_pop (unsigned char type, VOID_STAR ptr)
    return SLang_pop_complex (&z[0], &z[1]);
 }
 
-int _SLinit_slcomplex (void)
+int _pSLinit_slcomplex (void)
 {
    SLang_Class_Type *cl;
-   unsigned char *types;
+   SLtype *types;
 
    if (NULL == (cl = SLclass_allocate_class ("Complex_Type")))
      return -1;
@@ -1117,10 +1159,10 @@ int _SLinit_slcomplex (void)
 				     SLANG_CLASS_TYPE_VECTOR))
      return -1;
 
-   types = _SLarith_Arith_Types;
+   types = _pSLarith_Arith_Types;
    while (*types != SLANG_DOUBLE_TYPE)
      {
-	unsigned char t = *types++;
+	SLtype t = *types++;
 
 	if ((-1 == SLclass_add_binary_op (t, SLANG_COMPLEX_TYPE, generic_complex_binary, complex_binary_result))
 	    || (-1 == SLclass_add_binary_op (SLANG_COMPLEX_TYPE, t, complex_generic_binary, complex_binary_result))
