@@ -1,6 +1,6 @@
 /* SLang_read_line interface --- uses SLang tty stuff */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -39,6 +39,8 @@ typedef struct RL_History_Type
 struct _pSLrline_Type
 {
    RL_History_Type *root, *tail, *last;
+   RL_History_Type *saved_line;	       /* used when browsing the history */
+   char *name;			       /* the name of this object */
    unsigned char *buf;		       /* edit buffer */
    unsigned int buf_len;			       /* sizeof buffer */
    unsigned int point;			       /* current editing point */
@@ -66,14 +68,17 @@ struct _pSLrline_Type
    SLKeyMap_List_Type *keymap;
    int eof_char;
 
-   /* tty variables */
    unsigned int flags;		       /*  */
+
    int state;
 #define RLI_LINE_INVALID	0
 #define RLI_LINE_SET		1
 #define RLI_LINE_IN_PROGRESS	2
 #define RLI_LINE_READ		3
 
+   int is_modified;
+
+   /* tty variables */
    unsigned int (*getkey)(void);       /* getkey function -- required */
    void (*tt_goto_column)(int);
    void (*tt_insert)(char);
@@ -83,13 +88,59 @@ struct _pSLrline_Type
    VOID_STAR update_client_data;
    /* This function is only called when blinking matches */
    int (*input_pending)(int);
+   
+   SLang_Name_Type *completion_callback;
+   SLang_Name_Type *list_completions_callback;
 };
 
+   
+static SLang_Name_Type *Default_Completion_Callback;
+static SLang_Name_Type *Default_List_Completions_Callback;
+
 int SLang_Rline_Quit;
 
 static unsigned char Char_Widths[256];
 static void position_cursor (SLrline_Type *, int);
 
+static void free_history_item (RL_History_Type *h)
+{
+   if (h == NULL)
+     return;
+
+   if (h->buf != NULL)
+     SLang_create_slstring (h->buf);
+   SLfree ((char *)h);
+}
+
+static void free_history (RL_History_Type *h)
+{
+   while (h != NULL)
+     {
+	RL_History_Type *next = h->next;
+	free_history_item (h);
+	h = next;
+     }
+}
+
+static RL_History_Type *allocate_history (char *str, int point)
+{
+   RL_History_Type *h;
+
+   if (NULL == (h = (RL_History_Type *) SLcalloc (1, sizeof (RL_History_Type)))
+       || (NULL == (h->buf = SLang_create_slstring (str))))
+     {
+	SLfree ((char *)h);	       /* NULL ok */
+	return NULL;
+     }
+   
+   h->len = strlen (str);
+   if ((point < 0) || ((unsigned int)point > h->len))
+     point = h->len;
+   h->point = point;
+   return h;
+}
+
+
 static void rl_beep (void)
 {
    putc(7, stdout);
@@ -116,13 +167,13 @@ static int check_space (SLrline_Type *This_RLI, unsigned int dn)
 
 
 /* editing functions */
-static int rl_bol (SLrline_Type *This_RLI)
+int SLrline_bol (SLrline_Type *This_RLI)
 {
    This_RLI->point = 0;
    return 0;
 }
 
-static int rl_eol (SLrline_Type *This_RLI)
+int SLrline_eol (SLrline_Type *This_RLI)
 {
    This_RLI->point = This_RLI->len;
    return 0;
@@ -170,6 +221,30 @@ static int rl_left (SLrline_Type *This_RLI)
    return 0;
 }
 
+int SLrline_move (SLrline_Type *rli, int n)
+{
+   if (rli == NULL)
+     return -1;
+
+   if (n < 0)
+     {
+	n = -n;
+	while (n && rli->point)
+	  {
+	     (void) rl_left (rli);
+	     n--;
+	  }
+	return 0;
+     }
+   
+   while (n && (rli->point != rli->len))
+     {
+	(void) rl_right (rli);
+	n--;
+     }
+   return 0;
+}
+	
 int SLrline_ins (SLrline_Type *This_RLI, char *s, unsigned int n)
 {
    unsigned char *pmin;
@@ -191,6 +266,8 @@ int SLrline_ins (SLrline_Type *This_RLI, char *s, unsigned int n)
 
    This_RLI->len += n;
    This_RLI->point += n;
+   This_RLI->is_modified = 1;
+
    return n;
 }
 
@@ -229,6 +306,7 @@ int SLrline_del (SLrline_Type *This_RLI, unsigned int n)
      {
 	*p++ = *pn++;
      }
+   This_RLI->is_modified = 1;
    return 0;
 }
 
@@ -305,12 +383,13 @@ static int rl_deleol (SLrline_Type *This_RLI)
    if (This_RLI->point == This_RLI->len) return 0;
    *(This_RLI->buf + This_RLI->point) = 0;
    This_RLI->len = This_RLI->point;
-   return 1;
+   This_RLI->is_modified = 1;
+   return 0;
 }
 
 static int rl_delete_line (SLrline_Type *This_RLI)
 {
-   rl_bol (This_RLI);
+   (void) SLrline_bol (This_RLI);
    rl_deleol (This_RLI);
    return 0;
 }
@@ -325,6 +404,121 @@ static int rl_enter (SLrline_Type *This_RLI)
    return 0;
 }
 
+static int rl_complete (SLrline_Type *rli)
+{
+   char *line;
+   unsigned int i, n, nbytes;
+   char **strings, *str0, ch0;
+   int start_point, delta;
+   SLang_Array_Type *at;
+   SLang_Name_Type *completion_callback;
+   SLang_Name_Type *list_completions_callback;
+
+   if (NULL == (completion_callback = rli->completion_callback))
+     {
+	completion_callback = Default_Completion_Callback;
+	if (completion_callback == NULL)
+	  return SLrline_ins (rli, "\t", 1);
+     }
+   if (NULL == (list_completions_callback = rli->list_completions_callback))
+     list_completions_callback = Default_List_Completions_Callback;
+
+   if (NULL == (line = SLrline_get_line (rli)))
+     return -1;
+
+   if ((-1 == SLang_start_arg_list ())
+       || (-1 == SLang_push_string (line))
+       || (-1 == SLang_push_int (rli->point))
+       || (-1 == SLang_end_arg_list ())
+       || (-1 == SLexecute_function (completion_callback)))
+     {
+	SLfree (line);
+	return -1;
+     }
+
+   SLfree (line);
+
+   if (-1 == SLang_pop_int (&start_point))
+     return -1;
+   
+   if (start_point < 0)
+     start_point = 0;
+
+   if (-1 == SLang_pop_array_of_type (&at, SLANG_STRING_TYPE))
+     return -1;
+
+   strings = (char **) at->data;
+   n = at->num_elements;
+   
+   if (n == 0)
+     {
+	SLang_free_array (at);
+	return 0;
+     }
+
+   if ((n != 1) && (list_completions_callback != NULL))
+     {
+	if ((-1 == SLang_start_arg_list ())
+	    || (-1 == SLang_push_array (at, 0))
+	    || (-1 == SLang_end_arg_list ())
+	    || (-1 == SLexecute_function (list_completions_callback)))
+	  {
+	     SLang_free_array (at);
+	     return -1;
+	  }
+	(void) SLrline_redraw (rli);
+     }
+	
+   str0 = strings[0];
+   nbytes = 0;
+   while (0 != (ch0 = str0[nbytes]))
+     {
+	for (i = 1; i < n; i++)
+	  {
+	     char ch1 = strings[i][nbytes];
+	     if (ch0 != ch1)
+	       break;
+	  }
+	if (i != n)
+	  break;
+	nbytes++;
+     }
+
+   delta = start_point - rli->point;
+   if (delta < 0)
+     {
+	(void) SLrline_move (rli, delta);
+	delta = -delta;
+     }
+   (void) SLrline_del (rli, (unsigned int) delta);
+   (void) SLrline_ins (rli, str0, nbytes);
+
+   /* How should the completion be ended?
+    *   "foo/     -->  "foo/
+    *   "foo/bar  -->  "foo/bar"
+    *   "foo      -->  "foo"
+    *   foo       -->  fooSPACE
+    *   foo/bar   -->  fooSPACE
+    */
+   if ((n == 1) 
+       && nbytes && (str0[nbytes-1] != '/') && (str0[nbytes-1] != '\\'))
+     {
+	char qch = ' ';
+
+	if (start_point > 0)
+	  {
+	     ch0 = rli->buf[start_point-1];
+	     if ((ch0 == '"') || (ch0 == '\''))
+	       qch = ch0;
+	  }
+	if (qch != 0)
+	  (void) SLrline_ins (rli, &qch, 1);
+     }
+
+   SLang_free_array (at);
+   return 0;
+}
+
 static SLKeyMap_List_Type *RL_Keymap;
 
 static SLuchar_Type *compute_char_width (SLuchar_Type *b, SLuchar_Type *bmax, int utf8_mode, unsigned int *wp, SLwchar_Type *wchp, int *illegalp)
@@ -734,17 +928,24 @@ static void RLupdate (SLrline_Type *rli)
 
 void SLrline_redraw (SLrline_Type *rli)
 {
-   unsigned char *p;
-   unsigned char *pmax;
-   
    if (rli == NULL)
      return;
 
-   p = rli->new_upd;
-   pmax = p + rli->edit_width;
-   while (p < pmax) *p++ = ' ';
-   rli->new_upd_len = rli->edit_width;
-   really_update (rli, 0);
+   if (rli->update_hook != NULL)
+     {
+	(*rli->update_hook) (rli, "", "", 0, 0, rli->update_client_data);
+     }
+   else
+     {
+	unsigned char *p;
+	unsigned char *pmax;
+   
+	p = rli->new_upd;
+	pmax = p + rli->edit_width;
+	while (p < pmax) *p++ = ' ';
+	rli->new_upd_len = rli->edit_width;
+	really_update (rli, 0);
+     }
    RLupdate (rli);
 }
 
@@ -812,7 +1013,8 @@ static void blink_match (SLrline_Type *rli)
 	       {
 		  rli->point -= delta_column;
 		  RLupdate (rli);
-		  (*rli->input_pending)(10);
+		  if (rli->input_pending != NULL)
+		    (*rli->input_pending)(10);
 		  rli->point += delta_column;
 		  RLupdate (rli);
 		  break;
@@ -825,6 +1027,8 @@ static void blink_match (SLrline_Type *rli)
      }
 }
 
+static SLrline_Type *Active_Rline_Info = NULL;
+
 char *SLrline_read_line (SLrline_Type *rli, char *prompt, unsigned int *lenp)
 {
    unsigned char *p, *pmax;
@@ -878,11 +1082,16 @@ char *SLrline_read_line (SLrline_Type *rli, char *prompt, unsigned int *lenp)
    if (rli->update_hook == NULL)
      putc ('\r', stdout);
 
+   rli->is_modified = 0;
+   rli->last = NULL;
+
    RLupdate (rli);
 
    last_input_char = 0;
    while (1)
      {
+	SLrline_Type *save_rli = Active_Rline_Info;
+
 	key = SLang_do_key (RL_Keymap, (int (*)(void)) rli->getkey);
 
 	if ((key == NULL) || (key->f.f == NULL))
@@ -906,6 +1115,7 @@ char *SLrline_read_line (SLrline_Type *rli, char *prompt, unsigned int *lenp)
 	     last_input_char = rli->eof_char;
 	  }
 
+	Active_Rline_Info = rli;
 	if (key->type == SLKEY_F_INTRINSIC)
 	  {
 	     int (*func)(SLrline_Type *);
@@ -919,16 +1129,28 @@ char *SLrline_read_line (SLrline_Type *rli, char *prompt, unsigned int *lenp)
 		 && (rli->input_pending != NULL))
 	       blink_match (rli);
 	  }
-	
-	if ((SLang_Rline_Quit) || (_pSLang_Error == SL_USER_BREAK))
+	else if (key->type == SLKEY_F_SLANG)
 	  {
-	     if (_pSLang_Error == SL_USER_BREAK)
+	     (void) SLexecute_function (key->f.slang_fun);
+	     RLupdate (rli);
+	  }
+	Active_Rline_Info = save_rli;
+
+	if ((SLang_Rline_Quit) || _pSLang_Error)
+	  {
+	     if (_pSLang_Error)
 	       {
 		  rli->len = 0;
 	       }
 	     rli->buf[rli->len] = 0;
 	     rli->state = RLI_LINE_READ;
 	     *lenp = rli->len;
+	     
+	     free_history_item (rli->saved_line);
+	     rli->saved_line = NULL;
+	     
+	     if (_pSLang_Error)
+	       return NULL;
 
 	     return SLmake_nstring ((char *)rli->buf, rli->len);
 	  }
@@ -963,21 +1185,28 @@ static int rl_select_line (SLrline_Type *rli, RL_History_Type *p)
    strcpy ((char *) rli->buf, p->buf);
    rli->point = p->point;
    rli->len = len;
+   rli->is_modified = 0;
    return 0;
 }
 
-static int rl_next_line (SLrline_Type *This_RLI);
+static int rl_redraw (SLrline_Type *This_RLI)
+{
+   SLrline_redraw (This_RLI);
+   return 0;
+}
+
+
 static int rl_prev_line (SLrline_Type *This_RLI)
 {
    RL_History_Type *prev;
 
-   if (((This_RLI->last_fun != (FVOID_STAR) rl_prev_line)
-	&& (This_RLI->last_fun != (FVOID_STAR) rl_next_line))
+   if ((This_RLI->is_modified)
        || (This_RLI->last == NULL))
      {
 	prev = This_RLI->tail;
      }
-   else prev = This_RLI->last->prev;
+   else
+     prev = This_RLI->last->prev;
 
    if (prev == NULL)
      {
@@ -985,47 +1214,64 @@ static int rl_prev_line (SLrline_Type *This_RLI)
 	return 0;
      }
 
-   return rl_select_line (This_RLI, prev);
-}
+   if (prev == This_RLI->tail)
+     {
+	This_RLI->buf[This_RLI->len] = 0;
+	free_history_item (This_RLI->saved_line);
+	This_RLI->saved_line = allocate_history ((char *)This_RLI->buf, This_RLI->point);
+	if (This_RLI->saved_line == NULL)
+	  return -1;
+     }
 
-static int rl_redraw (SLrline_Type *This_RLI)
-{
-   SLrline_redraw (This_RLI);
-   return 1;
+   return rl_select_line (This_RLI, prev);
 }
 
 static int rl_next_line (SLrline_Type *This_RLI)
 {
    RL_History_Type *next;
 
-   if (((This_RLI->last_fun != (FVOID_STAR) rl_prev_line)
-	&& (This_RLI->last_fun != (FVOID_STAR) rl_next_line))
+   if ((This_RLI->is_modified)
        || (This_RLI->last == NULL))
-      {
-	 rl_beep ();
-	 return 0;
-      }
+     {
+	rl_beep ();
+	return 0;
+     }
 
    next = This_RLI->last->next;
 
    if (next == NULL)
      {
+	int status = 0;
+
+	if (This_RLI->saved_line != NULL)
+	  {	
+	     status = rl_select_line (This_RLI, This_RLI->saved_line);
+	     free_history_item (This_RLI->saved_line);
+	     This_RLI->saved_line = NULL;
+	     This_RLI->is_modified = 1;
+	     if (status == 0)
+	       return 0;
+	  }
 	This_RLI->len = This_RLI->point = 0;
 	*This_RLI->buf = 0;
 	This_RLI->last = NULL;
+	This_RLI->is_modified = 0;
+	return status;
      }
-   else rl_select_line (This_RLI, next);
-   return 1;
+   
+   return rl_select_line (This_RLI, next);
 }
 
 #define AKEY(name,func) {name,(int (*)(void))func}
 
 static SLKeymap_Function_Type SLReadLine_Functions[] =
 {
+   AKEY("kbd_quit",rl_abort),
+   AKEY("redraw",rl_redraw),
    AKEY("up", rl_prev_line),
    AKEY("down", rl_next_line),
-   AKEY("bol", rl_bol),
-   AKEY("eol", rl_eol),
+   AKEY("bol", SLrline_bol),
+   AKEY("eol", SLrline_eol),
    AKEY("right", rl_right),
    AKEY("left", rl_left),
    AKEY("self_insert", rl_self_insert),
@@ -1036,40 +1282,134 @@ static SLKeymap_Function_Type SLReadLine_Functions[] =
    AKEY("enter", rl_enter),
    AKEY("trim", rl_trim),
    AKEY("quoted_insert", rl_quote_insert),
+   AKEY("complete", rl_complete),
+   AKEY("self_insert", rl_self_insert),
    AKEY(NULL, NULL),
 };
 
-
-static void free_history (SLrline_Type *rli)
-{
-   RL_History_Type *h;
-   
-   h = rli->root;
-   while (h != NULL)
-     {
-	RL_History_Type *next = h->next;
-	if (h->buf != NULL)
-	  SLfree (h->buf);
-	SLfree ((char *)h);
-	h = next;
-     }
-}
-	     
 void SLrline_close (SLrline_Type *rli)
 {
    if (rli == NULL)
      return;
    
-   free_history (rli);
+   if (rli->name != NULL)
+     {
+	char hookname[1024];
+	SLrline_Type *arli = Active_Rline_Info;
+	Active_Rline_Info = rli;
+	SLsnprintf (hookname, sizeof(hookname), "%s_rline_close_hook", rli->name);
+	if (0 == SLang_run_hooks (hookname, 0))
+	  (void) SLang_run_hooks ("rline_close_hook", 1, rli->name);
+	Active_Rline_Info = arli;
+	SLang_free_slstring (rli->name);
+     }
+
+   free_history (rli->root);
+   free_history_item (rli->saved_line);
+   SLang_free_function (rli->list_completions_callback);
+   SLang_free_function (rli->completion_callback);
    SLfree (rli->prompt);
    SLfree ((char *)rli->buf);
    SLfree ((char *)rli);
 }
 
-SLrline_Type *SLrline_open (unsigned int width, unsigned int flags)
+static int init_keymap (void)
 {
    int ch;
    char simple[2];
+   SLkeymap_Type *km;
+
+   if (RL_Keymap != NULL)
+     return 0;
+
+   simple[1] = 0;
+   if (NULL == (km = SLang_create_keymap ("ReadLine", NULL)))
+     return -1;
+
+   km->functions = SLReadLine_Functions;
+
+   /* This breaks under some DEC ALPHA compilers (scary!) */
+#ifndef __DECC
+   for (ch = ' '; ch < 256; ch++)
+     {
+	simple[0] = (char) ch;
+	SLkm_define_key (simple, (FVOID_STAR) rl_self_insert, km);
+     }
+#else
+   ch = ' ';
+   while (1)
+     {
+	simple[0] = (char) ch;
+	SLkm_define_key (simple, (FVOID_STAR) rl_self_insert, km);
+	ch = ch + 1;
+	if (ch == 256) break;
+     }
+#endif				       /* NOT __DECC */
+   
+   simple[0] = SLang_Abort_Char;
+   SLkm_define_key (simple, (FVOID_STAR) rl_abort, km);
+#ifdef REAL_UNIX_SYSTEM
+   simple[0] = (char) 4;
+#else
+   simple[0] = (char) 26;
+#endif
+   SLkm_define_key (simple, (FVOID_STAR) rl_eof_insert, km);
+   
+#ifndef IBMPC_SYSTEM
+   SLkm_define_key  ("^[[A", (FVOID_STAR) rl_prev_line, km);
+   SLkm_define_key  ("^[[B", (FVOID_STAR) rl_next_line, km);
+   SLkm_define_key  ("^[[C", (FVOID_STAR) rl_right, km);
+   SLkm_define_key  ("^[[D", (FVOID_STAR) rl_left, km);
+   SLkm_define_key  ("^[OA", (FVOID_STAR) rl_prev_line, km);
+   SLkm_define_key  ("^[OB", (FVOID_STAR) rl_next_line, km);
+   SLkm_define_key  ("^[OC", (FVOID_STAR) rl_right, km);
+   SLkm_define_key  ("^[OD", (FVOID_STAR) rl_left, km);
+#else
+   SLkm_define_key  ("^@H", (FVOID_STAR) rl_prev_line, km);
+   SLkm_define_key  ("^@P", (FVOID_STAR) rl_next_line, km);
+   SLkm_define_key  ("^@M", (FVOID_STAR) rl_right, km);
+   SLkm_define_key  ("^@K", (FVOID_STAR) rl_left, km);
+   SLkm_define_key  ("^@S", (FVOID_STAR) rl_del, km);
+   SLkm_define_key  ("^@O", (FVOID_STAR) SLrline_eol, km);
+   SLkm_define_key  ("^@G", (FVOID_STAR) SLrline_bol, km);
+   
+   SLkm_define_key  ("\xE0H", (FVOID_STAR) rl_prev_line, km);
+   SLkm_define_key  ("\xE0P", (FVOID_STAR) rl_next_line, km);
+   SLkm_define_key  ("\xE0M", (FVOID_STAR) rl_right, km);
+   SLkm_define_key  ("\xE0K", (FVOID_STAR) rl_left, km);
+   SLkm_define_key  ("\xE0S", (FVOID_STAR) rl_del, km);
+   SLkm_define_key  ("\xE0O", (FVOID_STAR) SLrline_eol, km);
+   SLkm_define_key  ("\xE0G", (FVOID_STAR) SLrline_bol, km);
+#endif
+   SLkm_define_key  ("^C", (FVOID_STAR) rl_abort, km);
+   SLkm_define_key  ("^E", (FVOID_STAR) SLrline_eol, km);
+   SLkm_define_key  ("^G", (FVOID_STAR) rl_abort, km);
+   SLkm_define_key  ("^I", (FVOID_STAR) rl_complete, km);
+   SLkm_define_key  ("^A", (FVOID_STAR) SLrline_bol, km);
+   SLkm_define_key  ("\r", (FVOID_STAR) rl_enter, km);
+   SLkm_define_key  ("\n", (FVOID_STAR) rl_enter, km);
+   SLkm_define_key  ("^K", (FVOID_STAR) rl_deleol, km);
+   SLkm_define_key  ("^L", (FVOID_STAR) rl_deleol, km);
+   SLkm_define_key  ("^U", (FVOID_STAR) rl_delbol, km);
+   SLkm_define_key  ("^V", (FVOID_STAR) rl_del, km);
+   SLkm_define_key  ("^D", (FVOID_STAR) rl_del, km);
+   SLkm_define_key  ("^F", (FVOID_STAR) rl_right, km);
+   SLkm_define_key  ("^B", (FVOID_STAR) rl_left, km);
+   SLkm_define_key  ("^?", (FVOID_STAR) rl_bdel, km);
+   SLkm_define_key  ("^H", (FVOID_STAR) rl_bdel, km);
+   SLkm_define_key  ("^P", (FVOID_STAR) rl_prev_line, km);
+   SLkm_define_key  ("^N", (FVOID_STAR) rl_next_line, km);
+   SLkm_define_key  ("^R", (FVOID_STAR) rl_redraw, km);
+   SLkm_define_key  ("`", (FVOID_STAR) rl_quote_insert, km);
+   SLkm_define_key  ("\033\\", (FVOID_STAR) rl_trim, km);
+   if (_pSLang_Error)
+     return -1;
+   RL_Keymap = km;
+   return 0;
+}
+
+SLrline_Type *SLrline_open (unsigned int width, unsigned int flags)
+{
    SLrline_Type *rli;
 
    if (_pSLutf8_mode)
@@ -1110,100 +1450,18 @@ SLrline_Type *SLrline_open (unsigned int width, unsigned int flags)
 	if (rli->tt_goto_column == NULL) rli->tt_goto_column = ansi_goto_column;
      }
 
-   if (RL_Keymap == NULL)
+   if (-1 == init_keymap ())
      {
-	simple[1] = 0;
-	if (NULL == (RL_Keymap = SLang_create_keymap ("ReadLine", NULL)))
-	  {
-	     SLrline_close (rli);
-	     return NULL;
-	  }
-
-	RL_Keymap->functions = SLReadLine_Functions;
-
-	/* This breaks under some DEC ALPHA compilers (scary!) */
-#ifndef __DECC
-	for (ch = ' '; ch < 256; ch++)
-	  {
-	     simple[0] = (char) ch;
-	     SLkm_define_key (simple, (FVOID_STAR) rl_self_insert, RL_Keymap);
-	  }
-#else
-	ch = ' ';
-	while (1)
-	  {
-	     simple[0] = (char) ch;
-	     SLkm_define_key (simple, (FVOID_STAR) rl_self_insert, RL_Keymap);
-	     ch = ch + 1;
-	     if (ch == 256) break;
-	  }
-#endif				       /* NOT __DECC */
-
-	simple[0] = SLang_Abort_Char;
-	SLkm_define_key (simple, (FVOID_STAR) rl_abort, RL_Keymap);
-	simple[0] = (char) rli->eof_char;
-	SLkm_define_key (simple, (FVOID_STAR) rl_eof_insert, RL_Keymap);
-
-#ifndef IBMPC_SYSTEM
-	SLkm_define_key  ("^[[A", (FVOID_STAR) rl_prev_line, RL_Keymap);
-	SLkm_define_key  ("^[[B", (FVOID_STAR) rl_next_line, RL_Keymap);
-	SLkm_define_key  ("^[[C", (FVOID_STAR) rl_right, RL_Keymap);
-	SLkm_define_key  ("^[[D", (FVOID_STAR) rl_left, RL_Keymap);
-	SLkm_define_key  ("^[OA", (FVOID_STAR) rl_prev_line, RL_Keymap);
-	SLkm_define_key  ("^[OB", (FVOID_STAR) rl_next_line, RL_Keymap);
-	SLkm_define_key  ("^[OC", (FVOID_STAR) rl_right, RL_Keymap);
-	SLkm_define_key  ("^[OD", (FVOID_STAR) rl_left, RL_Keymap);
-#else
-	SLkm_define_key  ("^@H", (FVOID_STAR) rl_prev_line, RL_Keymap);
-	SLkm_define_key  ("^@P", (FVOID_STAR) rl_next_line, RL_Keymap);
-	SLkm_define_key  ("^@M", (FVOID_STAR) rl_right, RL_Keymap);
-	SLkm_define_key  ("^@K", (FVOID_STAR) rl_left, RL_Keymap);
-	SLkm_define_key  ("^@S", (FVOID_STAR) rl_del, RL_Keymap);
-	SLkm_define_key  ("^@O", (FVOID_STAR) rl_eol, RL_Keymap);
-	SLkm_define_key  ("^@G", (FVOID_STAR) rl_bol, RL_Keymap);
-
-	SLkm_define_key  ("\xE0H", (FVOID_STAR) rl_prev_line, RL_Keymap);
-	SLkm_define_key  ("\xE0P", (FVOID_STAR) rl_next_line, RL_Keymap);
-	SLkm_define_key  ("\xE0M", (FVOID_STAR) rl_right, RL_Keymap);
-	SLkm_define_key  ("\xE0K", (FVOID_STAR) rl_left, RL_Keymap);
-	SLkm_define_key  ("\xE0S", (FVOID_STAR) rl_del, RL_Keymap);
-	SLkm_define_key  ("\xE0O", (FVOID_STAR) rl_eol, RL_Keymap);
-	SLkm_define_key  ("\xE0G", (FVOID_STAR) rl_bol, RL_Keymap);
-#endif
-	SLkm_define_key  ("^C", (FVOID_STAR) rl_abort, RL_Keymap);
-	SLkm_define_key  ("^E", (FVOID_STAR) rl_eol, RL_Keymap);
-	SLkm_define_key  ("^G", (FVOID_STAR) rl_abort, RL_Keymap);
-	SLkm_define_key  ("^I", (FVOID_STAR) rl_self_insert, RL_Keymap);
-	SLkm_define_key  ("^A", (FVOID_STAR) rl_bol, RL_Keymap);
-	SLkm_define_key  ("\r", (FVOID_STAR) rl_enter, RL_Keymap);
-	SLkm_define_key  ("\n", (FVOID_STAR) rl_enter, RL_Keymap);
-	SLkm_define_key  ("^K", (FVOID_STAR) rl_deleol, RL_Keymap);
-	SLkm_define_key  ("^L", (FVOID_STAR) rl_deleol, RL_Keymap);
-	SLkm_define_key  ("^U", (FVOID_STAR) rl_delbol, RL_Keymap);
-	SLkm_define_key  ("^V", (FVOID_STAR) rl_del, RL_Keymap);
-	SLkm_define_key  ("^D", (FVOID_STAR) rl_del, RL_Keymap);
-	SLkm_define_key  ("^F", (FVOID_STAR) rl_right, RL_Keymap);
-	SLkm_define_key  ("^B", (FVOID_STAR) rl_left, RL_Keymap);
-	SLkm_define_key  ("^?", (FVOID_STAR) rl_bdel, RL_Keymap);
-	SLkm_define_key  ("^H", (FVOID_STAR) rl_bdel, RL_Keymap);
-	SLkm_define_key  ("^P", (FVOID_STAR) rl_prev_line, RL_Keymap);
-	SLkm_define_key  ("^N", (FVOID_STAR) rl_next_line, RL_Keymap);
-	SLkm_define_key  ("^R", (FVOID_STAR) rl_redraw, RL_Keymap);
-	SLkm_define_key  ("`", (FVOID_STAR) rl_quote_insert, RL_Keymap);
-	SLkm_define_key  ("\033\\", (FVOID_STAR) rl_trim, RL_Keymap);
-	if (_pSLang_Error)
-	  {
-	     SLrline_close (rli);
-	     return NULL;
-	  }
+	SLrline_close (rli);
+	return NULL;
      }
-
-   if (rli->keymap == NULL) rli->keymap = RL_Keymap;
+   rli->keymap = RL_Keymap;
    rli->old_upd = rli->upd_buf1;
    rli->new_upd = rli->upd_buf2;
 
    if (Char_Widths[0] == 0)
      {
+	int ch;
 	/* FIXME: This does not support UTF-8 */
 	for (ch = 0; ch < 32; ch++) Char_Widths[ch] = 2;
 	for (ch = 32; ch < 256; ch++) Char_Widths[ch] = 1;
@@ -1212,7 +1470,32 @@ SLrline_Type *SLrline_open (unsigned int width, unsigned int flags)
 	for (ch = 128; ch < 160; ch++) Char_Widths[ch] = 3;
 #endif
      }
+   return rli;
+}
 
+SLrline_Type *SLrline_open2 (char *name, unsigned int width, unsigned int flags)
+{
+   SLrline_Type *rli;
+   SLrline_Type *arli;
+   char hookname [1024];
+
+   if (NULL == (rli = SLrline_open (width, flags)))
+     return NULL;
+
+   if (NULL != rli->name)
+     SLang_free_slstring (rli->name);
+   if (NULL == (rli->name = SLang_create_slstring (name)))
+     {
+	SLrline_close (rli);
+	return NULL;
+     }
+
+   arli = Active_Rline_Info;
+   Active_Rline_Info = rli;
+   SLsnprintf (hookname, sizeof(hookname), "%s_rline_open_hook", name);
+   if (0 == SLang_run_hooks (hookname, 0))
+     (void) SLang_run_hooks ("rline_open_hook", 1, name);
+   Active_Rline_Info = arli;
    return rli;
 }
 
@@ -1223,12 +1506,7 @@ int SLrline_add_to_history (SLrline_Type *rli, char *hist)
    if ((rli == NULL) || (hist == NULL))
      return -1;
 
-   if (NULL == (h = (RL_History_Type *) SLcalloc (1, sizeof (RL_History_Type)))
-       || (NULL == (h->buf = SLmake_string (hist))))
-     {
-	SLfree ((char *)h);
-	return -1;
-     }
+   h = allocate_history (hist, -1);
 
    if (rli->root == NULL)
      rli->root = h;
@@ -1239,8 +1517,7 @@ int SLrline_add_to_history (SLrline_Type *rli, char *hist)
    h->prev = rli->tail;
    rli->tail = h;
    h->next = NULL;
-   
-   h->point = h->len = strlen (hist);
+
    return 0;
 }
    
@@ -1295,9 +1572,9 @@ int SLrline_set_point (SLrline_Type *rli, unsigned int point)
    if (rli->state == RLI_LINE_INVALID)
      return -1;
 
-   if (rli->len > point)
+   if (point > rli->len)
      point = rli->len;
-   
+
    rli->point = point;
    return 0;
 }
@@ -1393,3 +1670,377 @@ int SLrline_set_display_width (SLrline_Type *rli, unsigned int w)
    rli->edit_width = w;
    return 0;
 }
+
+/* SLang Interface */
+
+static void rline_ins_intrinsic (char *str)
+{
+   if (Active_Rline_Info != NULL)
+     (void) SLrline_ins (Active_Rline_Info, str, strlen (str));
+}
+
+static void rline_del_intrinsic (int *np)
+{
+   int n = *np;
+
+   if (Active_Rline_Info == NULL)
+     return;
+   
+   if (n < 0)
+     {
+	(void) SLrline_move (Active_Rline_Info, n);
+	n = -n;
+     }
+
+   (void) SLrline_del (Active_Rline_Info, (unsigned int) n);
+}
+
+static SLkeymap_Type *get_keymap (void)
+{
+   SLkeymap_Type *kmap;
+
+   if (Active_Rline_Info != NULL)
+     kmap = SLrline_get_keymap (Active_Rline_Info);
+   else
+     kmap = RL_Keymap;
+   
+   if (kmap != NULL)
+     return kmap;
+
+   SLang_verror (SL_APPLICATION_ERROR, "No keymap available for rline interface");
+   return NULL;
+}
+
+
+static void rline_setkey_intrinsic (char *keyseq)
+{
+   char *str;
+   SLkeymap_Type *kmap;
+
+   if (NULL == (kmap = get_keymap ()))
+     return;
+
+   if (SLang_peek_at_stack () == SLANG_REF_TYPE)
+     {
+	SLang_Name_Type *nt;
+	
+	if (NULL == (nt = SLang_pop_function ()))
+	  return;
+
+	(void) SLkm_define_slkey (keyseq, nt, kmap);
+	return;
+     }
+   
+   if (-1 == SLang_pop_slstring (&str))
+     return;
+   
+   (void) SLang_define_key (keyseq, str, kmap);
+   SLang_free_slstring (str);
+}
+
+static void rline_unsetkey_intrinsic (char *keyseq)
+{
+   SLkeymap_Type *kmap;
+   
+   if (NULL != (kmap = get_keymap ()))
+     SLang_undefine_key (keyseq, kmap);
+}
+
+static int rline_get_point_intrinsic (void)
+{
+   unsigned int p;
+
+   if ((Active_Rline_Info == NULL) 
+       || (-1 == SLrline_get_point (Active_Rline_Info, &p)))
+     return 0;
+
+   return (int) p;
+}
+
+static void rline_set_point_intrinsic (int *pp)
+{
+   int p;
+   SLrline_Type *rli;
+
+   if (NULL == (rli = Active_Rline_Info))
+     return;
+   
+   p = *pp;
+   if (p < 0)
+     {
+	p = (int)rli->len + (p + 1);
+	if (p < 0)
+	  p = 0;
+     }
+
+   if ((unsigned int)p > rli->len)
+     p = (int)rli->len;
+	
+   (void) SLrline_set_point (rli, (unsigned int) p);
+}
+
+static void rline_call_intrinsic (char *fun)
+{
+   int (*f)(SLrline_Type *);
+
+   if (Active_Rline_Info == NULL)
+     return;
+   
+   if (NULL == (f = (int (*)(SLrline_Type *)) (SLang_find_key_function(fun, Active_Rline_Info->keymap))))
+     {
+	SLang_verror (SL_UndefinedName_Error, "rline internal function %s does not exist", fun);
+	return;
+     }
+
+   (void) (*f)(Active_Rline_Info);
+   Active_Rline_Info->last_fun = (FVOID_STAR) f;
+}
+
+static void rline_get_line_intrinsic (void)
+{
+   char *s;
+
+   if ((Active_Rline_Info == NULL)
+       || (NULL == (s = SLrline_get_line (Active_Rline_Info))))
+     {
+	(void) SLang_push_string ("");
+	return;
+     }
+   (void) SLang_push_malloced_string (s);
+}
+
+static void rline_set_line_intrinsic (char *str)
+{
+   if (Active_Rline_Info == NULL)
+     return;
+
+   (void) SLrline_set_line (Active_Rline_Info, str);
+}
+
+static void rline_set_completion_callback (void)
+{
+   SLang_Name_Type *nt;
+
+   if (NULL == (nt = SLang_pop_function ()))
+     return;
+   
+   if (Active_Rline_Info == NULL)
+     {
+	SLang_free_function (Default_Completion_Callback);
+	Default_Completion_Callback = nt;
+	return;
+     }
+   SLang_free_function (Active_Rline_Info->completion_callback);
+   Active_Rline_Info->completion_callback = nt;
+}
+
+static void rline_set_list_completions_callback (void)
+{
+   SLang_Name_Type *nt;
+
+   if (NULL == (nt = SLang_pop_function ()))
+     return;
+   
+   if (Active_Rline_Info == NULL)
+     {
+	SLang_free_function (Default_List_Completions_Callback);
+	Default_List_Completions_Callback = nt;
+	return;
+     }
+   SLang_free_function (Active_Rline_Info->list_completions_callback);
+   Active_Rline_Info->list_completions_callback = nt;
+}
+
+static int rline_input_pending_intrinsic (int *tsecsp)
+{
+   int tsecs = *tsecsp;
+   if (tsecs < 0)
+     tsecs = 0;
+
+   if (Active_Rline_Info == NULL)
+     return 0;
+   
+   if (Active_Rline_Info->input_pending == NULL)
+     return 1;
+   
+   return Active_Rline_Info->input_pending (tsecs);
+}
+
+static int rline_getkey_intrinsic (void)
+{
+   if (Active_Rline_Info == NULL)
+     return -1;
+   
+   return (int) Active_Rline_Info->getkey ();
+}
+   
+static int rline_bolp_intrinsic (void)
+{
+   if (Active_Rline_Info == NULL)
+     return 0;
+   return Active_Rline_Info->point == 0;
+}
+
+static int rline_eolp_intrinsic (void)
+{
+   if (Active_Rline_Info == NULL)
+     return 0;
+   return Active_Rline_Info->point == Active_Rline_Info->len;
+}
+
+static int rline_get_edit_width_intrinsic (void)
+{
+   if (Active_Rline_Info == NULL)
+     return 80;
+   return Active_Rline_Info->edit_width;
+}
+
+static void rline_get_history_intrinsic (void)
+{
+   SLindex_Type i, num;
+   RL_History_Type *h;
+   char **data;
+   SLang_Array_Type *at;
+   
+   if (Active_Rline_Info == NULL)
+     {
+	SLang_push_null ();
+	return;
+     }
+   
+   num = 0;
+   h = Active_Rline_Info->root;
+   while (h != NULL)
+     {
+	h = h->next;
+	num++;
+     }
+   if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, NULL, &num, 1)))
+     return;
+
+   data = (char **)at->data;
+   h = Active_Rline_Info->root;
+   for (i = 0; i < num; i++)
+     {
+	if (NULL == (data[i] = SLang_create_slstring (h->buf)))
+	  {
+	     SLang_free_array (at);
+	     return;
+	  }
+	h = h->next;
+     }
+   
+   (void) SLang_push_array (at, 1);
+}
+
+static void rline_set_history_intrinsic (void)
+{
+   int i, num;
+   char **data;
+   SLrline_Type *rli;
+   SLang_Array_Type *at;
+
+   if (-1 == SLang_pop_array_of_type (&at, SLANG_STRING_TYPE))
+     return;
+
+   if (NULL == (rli = Active_Rline_Info))
+     {
+	SLang_free_array (at);
+	return;
+     }
+   
+   free_history (rli->root);
+   rli->tail = rli->root = rli->last = NULL;
+
+   data = (char **)at->data;
+   num = at->num_elements;
+   for (i = 0; i < num; i++)
+     {
+	if (-1 == SLrline_add_to_history (rli, data[i]))
+	  break;
+     }
+   SLang_free_array (at);
+}
+
+static SLang_Intrin_Fun_Type Intrinsics [] =
+{
+   MAKE_INTRINSIC_S("rline_call", rline_call_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_S("rline_ins", rline_ins_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_I("rline_del", rline_del_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_0("rline_bolp", rline_bolp_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_0("rline_eolp", rline_eolp_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_0("rline_get_point", rline_get_point_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_I("rline_set_point", rline_set_point_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_0("rline_get_edit_width", rline_get_edit_width_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_0("rline_get_line", rline_get_line_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_S("rline_set_line", rline_set_line_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_S("rline_setkey", rline_setkey_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_S("rline_unsetkey", rline_unsetkey_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_0("rline_getkey", rline_getkey_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_0("rline_set_history", rline_set_history_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_0("rline_get_history", rline_get_history_intrinsic, VOID_TYPE),
+   MAKE_INTRINSIC_I("rline_input_pending", rline_input_pending_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_0("rline_set_completion_callback", rline_set_completion_callback, VOID_TYPE),
+   MAKE_INTRINSIC_0("rline_set_list_completions_callback", rline_set_list_completions_callback, VOID_TYPE),
+   SLANG_END_INTRIN_FUN_TABLE
+};
+
+int SLrline_init (char *appname, char *user_initfile, char *sys_initfile)
+{
+#ifdef __WIN32__
+   char *home_dir = getenv ("USERPROFILE");
+#else
+# ifdef VMS
+   char *home_dir = "SYS$LOGIN:";
+# else
+   char *home_dir = getenv ("HOME");
+# endif
+#endif
+   char *file = NULL;
+   int status;
+   static char *appname_malloced;
+
+   if (sys_initfile == NULL)
+     sys_initfile = SLRLINE_SYS_INIT_FILE;
+   if (user_initfile == NULL)
+     user_initfile = SLRLINE_USER_INIT_FILE;
+   
+   if (appname == NULL)
+     appname = "Unknown";
+
+   if (NULL == (appname_malloced = SLmake_string (appname)))
+     return -1;
+
+   if (-1 == SLadd_intrinsic_variable ("__RL_APP__", &appname_malloced, SLANG_STRING_TYPE, 1))
+     return -1;
+
+   if (-1 == SLadd_intrin_fun_table (Intrinsics, NULL))
+     return -1;
+
+   if (-1 == init_keymap ())
+     return -1;
+
+   if (user_initfile != NULL)
+     {
+	file = SLpath_find_file_in_path (home_dir, user_initfile);
+	if (file != NULL)
+	  {
+	     status = SLns_load_file (file, NULL);
+	     SLfree (file);
+	     return status;
+	  }
+     }
+
+   if (sys_initfile != NULL)
+     {
+	file = _pSLpath_find_file (sys_initfile, 0);
+	if (file != NULL)
+	  {
+	     status = SLns_load_file (file, NULL);
+	     SLang_free_slstring (file);
+	     return status;
+	  }
+     }
+
+   return 0;
+}
