@@ -1,10 +1,24 @@
 /* sscanf function for S-Lang */
-/* Copyright (c) 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 #include <ctype.h>
@@ -137,15 +151,52 @@ static int parse_double (char **sp, char *smax, double *d)
    int has_leading_zeros;
    char *start_pos, *sign_pos;
    char *b, *bmax;
+   char ch;
 
    start_pos = *sp;
    s = get_sign (start_pos, smax, &sign);
    if (s >= smax)
      {
-	errno = _SLerrno_errno = EINVAL;
+	errno = _pSLerrno_errno = EINVAL;
 	return 0;
      }
 
+   ch = *s|0x20;
+   if ((ch == 'n') || (ch == 'i'))
+     {
+	if (s + 3 <= smax)
+	  {
+	     if (ch == 'n')
+	       {
+		  if (((s[1]|0x20) == 'a') && (s[2] == 'n'))
+		    {
+		       *sp = s + 3;
+		       *d = _pSLang_NaN;
+		       return 1;
+		    }
+		  *sp = start_pos;
+		  errno = _pSLerrno_errno = EINVAL;
+		  return 0;
+	       }
+	     if (((s[1] | 0x20) == 'n') && ((s[2] | 0x20) == 'f'))
+	       {
+		  /* check for infinity */
+		  if ((s + 8 <= smax)
+		      && (((s[3]|0x20)=='i')&&((s[4]|0x20)=='n')&&((s[5]|0x20)=='i')
+		     && ((s[6]|0x20)=='t')&&((s[7]|0x20)=='y')))
+		    *sp = s + 8;
+		  else
+		    *sp = s + 3;
+		  *d = _pSLang_Inf;
+		  return 1;
+	       }
+	  }
+	*sp = start_pos;
+	errno = _pSLerrno_errno = EINVAL;
+	return 0;
+     }
+
+   
    /* Prepare the buffer that will be passed to strtod */
    /* Allow the exponent to be 5 significant digits: E+xxxxx\0 */
    bmax = buf + (sizeof (buf) - 8);
@@ -354,8 +405,9 @@ static int parse_range (char **sp, char *smax, char **fp, char **str)
    return 1;
 }
 
-  
-int _SLang_sscanf (void)
+
+/* FIXME: This function does not handle LONG_LONG */
+int _pSLang_sscanf (void)
 {
    int num;
    unsigned int num_refs;
@@ -424,7 +476,10 @@ int _SLang_sscanf (void)
 
 	if (isspace (chf))
 	  {
-	     s = _SLskip_whitespace (s);
+	     char *s1 = _pSLskip_whitespace (s);
+	     if (s1 == s)
+	       break;
+	     s = s1;
 	     continue;
 	  }
 	
@@ -477,7 +532,11 @@ int _SLang_sscanf (void)
 	status = -1;
 
 	if ((chf != 'c') && (chf != '['))
-	  s = _SLskip_whitespace (s);
+	  {
+	     s = _pSLskip_whitespace (s);
+	     if (*s == 0)
+	       break;
+	  }
 
 	if (has_width)
 	  {
@@ -652,7 +711,7 @@ int _SLang_sscanf (void)
 	     goto return_error;
 	  }
 	
-	if (-1 == _SLang_deref_assign (ref))
+	if (-1 == _pSLang_deref_assign (ref))
 	  {
 	     SLang_free_ref (ref);
 	     goto return_error;
@@ -687,11 +746,11 @@ int _SLang_sscanf (void)
 extern double atof ();
 #endif
 
-double _SLang_atof (char *s)
+double _pSLang_atof (char *s)
 {
    double x;
 
-   s = _SLskip_whitespace (s);
+   s = _pSLskip_whitespace (s);
    errno = 0;
 
    if (1 != parse_double (&s, s + strlen (s), &x))
@@ -701,12 +760,12 @@ double _SLang_atof (char *s)
 	    || (0 == strcmp ("Inf", s)))
 	  return atof (s);	       /* let this deal with it */
 #ifdef EINVAL
-	errno = _SLerrno_errno = EINVAL;
+	errno = _pSLerrno_errno = EINVAL;
 #endif
 	return 0.0;
      }
    if (errno) 
-     _SLerrno_errno = errno;
+     _pSLerrno_errno = errno;
    return x;
 }
 #endif
