@@ -1,6 +1,6 @@
 /* SLang Screen management routines */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -709,21 +709,33 @@ void SLsmg_write_chars (unsigned char *u, unsigned char *umax)
 		  continue;
 	       }
 
-	     if (col + 2 > max_col)
-	       {
-		  ADD_CHAR_OR_BREAK('>');
-		  break;
-	       }
-	     
-	     if (col == start_col - 1)
+	     if (col + 1 == start_col)
 	       {
 		  /* double width character is clipped at left part of screen.
-		   * So, display right edge as a space */
-		  col++;
+		   * So, display right edge as a space 
+		   */
+		  col++;	       /* left edge */
 		  ADD_CHAR_OR_BREAK('<');
 		  continue;
 	       }
+	     
+	     if (i != 0) 	       /* finish active cell */
+	       NEXT_CHAR_CELL;
 
+	     if (last_was_double_width)
+	       {
+		  /* and right half of the cell */
+		  last_was_double_width = 0;
+		  NEXT_CHAR_CELL;
+	       }
+
+	     if (col + 2 > max_col)
+	       {
+		  /* right side of character gets clipped */
+		  ADD_TO_CHAR_CELL('>');
+		  col++;
+		  break;
+	       }
 	     ADD_CHAR_OR_BREAK(wc);
 	     last_was_double_width = 1;
 	     continue;
@@ -808,34 +820,6 @@ void SLsmg_write_string (char *str)
 		      (unsigned char *)str + strlen (str));
 }
 
-void SLsmg_write_nstring (char *str, unsigned int n)
-{
-   unsigned int width;
-   unsigned char *blank = (unsigned char *)" ";
-   unsigned char *u = (unsigned char *)str;
-
-   /* Avoid a problem if a user accidently passes a negative value */
-   if ((int) n < 0)
-     return;
-
-   if (u == NULL) width = 0;
-   else
-     {
-	unsigned char *umax;
-	
-	width = strlen ((char *)u);
-	if (UTF8_Mode)
-	  umax = SLutf8_skip_chars (u, u+width, n, &width, 0);
-	else 
-	  {
-	     if (width > n) width = n;
-	     umax = u + width;
-	  }
-	SLsmg_write_chars (u, umax);
-     }
-
-   while (width++ < n) SLsmg_write_chars (blank, blank+1);
-}
 
 void SLsmg_write_wrapped_string (SLuchar_Type *u, int r, int c,
 				 unsigned int dr, unsigned int dc,
@@ -844,15 +828,26 @@ void SLsmg_write_wrapped_string (SLuchar_Type *u, int r, int c,
    int maxc = (int) dc;
    unsigned char *p, *pmax;
    int utf8_mode = UTF8_Mode;
+   unsigned char display_8bit = (unsigned char) SLsmg_Display_Eight_Bit;
+
+   if (utf8_mode)
+     display_8bit = 0xA0;
 
    if ((dr == 0) || (dc == 0)) return;
+   if (u == NULL)
+     u = (unsigned char *)"";
+
    p = u;
    pmax = u + strlen ((char *)u);
    
    dc = 0;
    while (1)
      {
+	SLwchar_Type wc;
+	unsigned int nconsumed;
 	unsigned char ch = *p;
+	unsigned int ddc;
+
 	if ((ch == 0) || (ch == '\n'))
 	  {
 	     int diff;
@@ -876,27 +871,82 @@ void SLsmg_write_wrapped_string (SLuchar_Type *u, int r, int c,
 	     continue;
 	  }
 
-	if ((int) dc == maxc)
-	  {
-	     SLsmg_gotorc (r, c);
-	     SLsmg_write_chars (u, p);
-	     if (dr == 1) break;
+	/* If the width of the characters buffered so far extend to or beyond
+	 * the width of the box, then write them out and goto the
+	 * next line. Note that dc > maxc if the displayable width of
+	 * the last character to be written is greater than the width
+	 * of the box.  This will be the case if (maxc<ddc) -- see below
+	 */
+	if ((int) dc >= maxc)
+	  goto write_chars_and_reset;
 
-	     r++;
-	     dc = 0;
-	     dr--;
-	     u = p;
+	nconsumed = 1;
+	if (ch < 0x80)
+	  {
+	     p++;
+	     if ((ch >= 0x20) && (ch != 0x7F))
+	       {
+		  dc++;
+		  continue;
+	       }
+	     /* Otherwise display as ^X */
+	     dc += 2;
 	     continue;
 	  }
 
-	dc++;
-	if (utf8_mode)
-	  p = SLutf8_skip_chars (p, pmax, 1, NULL, 0);
+	nconsumed = 1;
+	if ((utf8_mode == 0)
+	    || (NULL == SLutf8_decode (p, pmax, &wc, &nconsumed)))
+	  {	     
+	     if ((utf8_mode == 0)
+		 && (display_8bit && (*p >= display_8bit)))
+	       {
+		  dc++;
+		  p += nconsumed;
+		  continue;
+	       }
+
+	     ddc = 4*nconsumed;      /* <XX> */
+	  }
+	else if (wc < (SLwchar_Type)display_8bit)
+	  ddc = 4;		       /* displays as <XX> */
 	else
-	  p++;
+	  ddc = SLwchar_wcwidth (wc);
+
+	dc += ddc;
+	if (((int)dc > maxc) && (maxc > (int)ddc))
+	  {
+	     dc -= ddc;
+	     goto write_chars_and_reset;
+	  }
+	p += nconsumed;
+	continue;
+
+	write_chars_and_reset:	
+	SLsmg_gotorc (r, c);
+	SLsmg_write_chars (u, p);
+	while ((int)dc < maxc)
+	  {
+	     SLsmg_write_char (' ');
+	     dc++;
+	  }
+	if (dr == 1) break;
+	r++;
+	dc = 0;
+	dr--;
+	u = p;
      }
 }
 
+void SLsmg_write_nstring (char *str, unsigned int n)
+{
+   /* Avoid a problem if a user accidently passes a negative value */
+   if ((int) n < 0)
+     return;
+   
+   SLsmg_write_wrapped_string ((unsigned char *)str, This_Row, This_Col, 1, n, 1);
+}
+
 int SLsmg_Tab_Width = 8;
 
 /* Minimum value for which eight bit char is displayed as is. */
@@ -950,13 +1000,19 @@ static void do_copy (SLsmg_Char_Type *a, SLsmg_Char_Type *b)
 int SLsmg_Scroll_Hash_Border = 0;
 static unsigned long compute_hash (SLsmg_Char_Type *c, unsigned int n)
 {
-   SLsmg_Char_Type *csave, *cmax;
+   SLsmg_Char_Type *cmin, *cmax;
    int is_blank = 2;
 
-   c += SLsmg_Scroll_Hash_Border;
-   csave = c;
-   cmax = c + (n - SLsmg_Scroll_Hash_Border);
+   cmin = c;
+   cmax = c + n;
 
+   if (SLsmg_Scroll_Hash_Border > 0)
+     {
+	cmax -= SLsmg_Scroll_Hash_Border;
+	cmin += SLsmg_Scroll_Hash_Border;
+     }
+
+   c = cmin;
    while ((c < cmax) && is_blank)
      {
 	if ((c->wchars[0] != 32) || (c->nchars != 1))
@@ -965,7 +1021,7 @@ static unsigned long compute_hash (SLsmg_Char_Type *c, unsigned int n)
      }
    if (is_blank) return 0;
 
-   return _pSLstring_hash ((unsigned char *)csave, (unsigned char *)cmax);
+   return _pSLstring_hash ((unsigned char *)cmin, (unsigned char *)cmax);
 }
 
 static unsigned long Blank_Hash;
