@@ -1,7 +1,7 @@
 /* -*- mode: C; mode: fold; -*- */
 /* string manipulation functions for S-Lang. */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -209,13 +209,18 @@ static void strcat_cmd (void) /*{{{*/
 
 /*}}}*/
 
+_INLINE_
 static int _pSLang_push_nstring (char *a, unsigned int len)
 {
    a = SLang_create_nslstring (a, len);
    if (a == NULL)
      return -1;
-   
-   return _pSLang_push_slstring (a);
+
+   if (0 == SLclass_push_ptr_obj (SLANG_STRING_TYPE, (VOID_STAR)a))
+     return 0;
+
+   SLang_free_slstring (a);
+   return -1;
 }
 
 
@@ -1265,8 +1270,6 @@ static int strbytelen_cmd (SLstr_Type *s)
 /*}}}*/
 
 
-/* sprintf functionality for S-Lang */
-
 static char *SLdo_sprintf (char *fmt) /*{{{*/
 {
    register char *p = fmt, ch;
@@ -1287,6 +1290,7 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
    int use_double;
    double x;
 #endif
+   unsigned char uch;
    int use_long = 0;
 
    while (1)
@@ -1468,13 +1472,18 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 		  ch = 's';
 		  str = (char *)utf8_buf;
 		  use_string = 1;
-		  break;
 	       }
-#if 0
-	     guess_size = 1;
+	     break;
+	     
+	   case 'b':
 	     use_long = 0;
-	     /* drop */
-#endif
+	     guess_size = 1;
+	     if (-1 == SLang_pop_uchar (&uch))
+	       return out;
+	     int_var = (int) uch;
+	     ch = 'c';
+	     break;
+
 	   case 'd':
 	   case 'i':
 	   case 'o':
@@ -1509,6 +1518,18 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 	     if (SLang_pop_double(&x)) return (out);
 	     use_double = 1;
 	     guess_size = 256;
+	     if (fabs(x) > 1e38)
+	       {
+		  if (0 == _pSLmath_isinf (x))
+		    {
+		       double expon = log10 (fabs(x));
+		       if (expon > (double) 0xFFFF)
+			 ch = 'E';
+		       else
+			 guess_size += (unsigned int) expon;
+		    }
+		  else ch = 'E';
+	       }
 	     use_long = 0;
 	     break;
 #endif
@@ -2015,13 +2036,24 @@ static void glob_to_regexp (char *glob)
    char ch;
 
    len = _pSLstring_bytelen (glob);
-   pat = SLmalloc (2*len + 3);
+   pat = SLmalloc (2*len + 8);
    if (pat == NULL)
      return;
-   
+
    p = pat;
    *p++ = '^';
 
+   /* If the first character of a file is '.', it must be explicitly matched. */
+   if ((*glob == '?') || (*glob == '*'))
+     {
+	*p++ = '[';
+	*p++ = '^';
+	*p++ = '.';
+	*p++ = ']';
+	if (*glob == '?')
+	  glob++;
+     }
+
    while (0 != (ch = *glob++))
      {
 	if ((ch == '.') || (ch == '$') || (ch == '+') || (ch == '\\'))
@@ -2047,25 +2079,33 @@ static void glob_to_regexp (char *glob)
 	     continue;
 	  }
 	
+	/* Otherwise ch = '[' */
 	if (*glob != 0)
 	  {
 	     char *g = glob;
-	     
+	     int is_complement = 0;
+
 	     ch = *g;
-	     if (ch == '^')
+	     if ((ch == '^') || (ch == '!'))
 	       {
+		  is_complement = 1;
 		  g++;
 		  ch = *g;
 	       }
 	     if (ch == ']')
 	       g++;
-	     
+
 	     while (((ch = *g) != 0) && (ch != ']'))
 	       g++;
 	     
 	     if (ch == ']')
 	       {
 		  *p++ = '[';
+		  if (is_complement)
+		    {
+		       *p++ = '^';
+		       glob++;
+		    }
 		  while (glob <= g)
 		    *p++ = *glob++;
 		  
