@@ -1,6 +1,6 @@
 /* Structure type implementation */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -27,7 +27,7 @@ USA.
 #include "_slang.h"
 
 
-void _pSLstruct_delete_struct (_pSLang_Struct_Type *s)
+void SLang_free_struct (_pSLang_Struct_Type *s)
 {
    _pSLstruct_Field_Type *field, *field_max;
 
@@ -117,12 +117,15 @@ static int push_struct_of_type (SLtype type, _pSLang_Struct_Type *s)
    return -1;
 }
 
-int _pSLang_push_struct (_pSLang_Struct_Type *s)
+int SLang_push_struct (_pSLang_Struct_Type *s)
 {
+   if (s == NULL)
+     return SLang_push_null ();
+
    return push_struct_of_type (SLANG_STRUCT_TYPE, s);
 }
 
-int _pSLang_pop_struct (_pSLang_Struct_Type **sp)
+int SLang_pop_struct (_pSLang_Struct_Type **sp)
 {
    SLang_Object_Type obj;
    SLang_Class_Type *cl;
@@ -153,7 +156,7 @@ int _pSLang_pop_struct (_pSLang_Struct_Type **sp)
 static void struct_destroy (SLtype type, VOID_STAR vs)
 {
    (void) type;
-   _pSLstruct_delete_struct (*(_pSLang_Struct_Type **) vs);
+   SLang_free_struct (*(_pSLang_Struct_Type **) vs);
 }
 
 static int struct_push (SLtype type, VOID_STAR ptr)
@@ -180,6 +183,17 @@ static _pSLstruct_Field_Type *find_field (_pSLang_Struct_Type *s, char *name)
    return NULL;
 }
 
+/* This function is used by the qualifier-code */
+SLang_Object_Type *_pSLstruct_get_field_value (SLang_Struct_Type *s, char *name)
+{
+   _pSLstruct_Field_Type *f = find_field (s, name);
+
+   if (f == NULL)
+     return NULL;
+
+   return &f->obj;
+}
+
 static _pSLstruct_Field_Type *pop_field (_pSLang_Struct_Type *s, char *name,
 					_pSLstruct_Field_Type *(*find)(_pSLang_Struct_Type *, char *))
 {
@@ -244,7 +258,7 @@ static _pSLang_Struct_Type *
    return s;
 
    return_error:
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return NULL;
 }
 
@@ -258,10 +272,10 @@ int SLstruct_create_struct (unsigned int nfields,
    if (NULL == (s = create_struct (nfields, field_names, field_types, field_values)))
      return -1;
 
-   if (0 == _pSLang_push_struct (s))
+   if (0 == SLang_push_struct (s))
      return 0;
 
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return -1;
 }
 
@@ -292,7 +306,7 @@ static _pSLang_Struct_Type *struct_from_struct_fields (int nfields)
 	nfields--;
 	if (-1 == SLang_pop_slstring (&name))
 	  {
-	     _pSLstruct_delete_struct (s);
+	     SLang_free_struct (s);
 	     return NULL;
 	  }
 
@@ -306,7 +320,7 @@ static _pSLang_Struct_Type *struct_from_struct_fields (int nfields)
 	     SLang_verror (SL_DuplicateDefinition_Error,
 			   "Field %s used more than once in the struct",
 			   name);
-	     _pSLstruct_delete_struct (s);
+	     SLang_free_struct (s);
 	     return NULL;
 	  }
      }
@@ -325,14 +339,69 @@ int _pSLstruct_define_struct (void)
    if (NULL == (s = struct_from_struct_fields (nfields)))
      return -1;
 
-   if (-1 == _pSLang_push_struct (s))
+   if (-1 == SLang_push_struct (s))
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return -1;
      }
    return 0;
 }
 
+static int pop_to_struct_field (_pSLang_Struct_Type *s, char *name)
+{
+   _pSLstruct_Field_Type *f;
+   SLang_Object_Type obj;
+   
+   if ((NULL == (f = pop_field (s, name, find_field)))
+       || (-1 == SLang_pop (&obj)))
+     return -1;
+
+   SLang_free_object (&f->obj);
+   f->obj = obj;
+   
+   return 0;
+}
+
+/* This function is used for structure definitions with embedded assignments */
+int _pSLstruct_define_struct2 (void)
+{
+   _pSLang_Struct_Type *s;
+   int nfields;
+   int nassigns;
+
+   if (-1 == SLang_pop_integer (&nassigns))
+     return -1;
+
+   if (-1 == SLang_pop_integer (&nfields))
+     return -1;
+
+   if (NULL == (s = struct_from_struct_fields (nfields)))
+     return -1;
+
+   while (nassigns > 0)
+     {
+	char *name;
+	if (-1 == SLang_pop_slstring (&name))
+	  goto return_error;
+	
+	if (-1 == pop_to_struct_field (s, name))
+	  {
+	     SLang_free_slstring (name);
+	     goto return_error;
+	  }
+	SLang_free_slstring (name);
+	nassigns--;
+     }
+
+   if (0 == SLang_push_struct (s))
+     return 0;
+   
+   return_error:
+   
+   SLang_free_struct (s);
+   return -1;
+}
+
 static int init_struct_with_user_methods (SLtype, _pSLang_Struct_Type *);
 /* Simply make a struct that contains the same fields as struct s.  Do not
  * duplicate the field values.
@@ -354,7 +423,7 @@ static _pSLang_Struct_Type *make_struct_shell (_pSLang_Struct_Type *s, SLtype ty
      {
 	if (NULL == (new_f[i].name = SLang_create_slstring (old_f[i].name)))
 	  {
-	     _pSLstruct_delete_struct (new_s);
+	     SLang_free_struct (new_s);
 	     return NULL;
 	  }
      }
@@ -390,7 +459,7 @@ typedefed_struct_datatype_deref (SLtype type)
 
    if (-1 == push_struct_of_type (type, s))
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return -1;
      }
 
@@ -421,7 +490,7 @@ static _pSLang_Struct_Type *duplicate_struct (_pSLang_Struct_Type *s, SLtype typ
 	     if ((-1 == _pSLpush_slang_obj (obj))
 		 || (-1 == SLang_pop (&new_f->obj)))
 	       {
-		  _pSLstruct_delete_struct (new_s);
+		  SLang_free_struct (new_s);
 		  return NULL;
 	       }
 	  }
@@ -441,7 +510,7 @@ static int struct_dereference (SLtype type, VOID_STAR addr)
 
    if (-1 == push_struct_of_type (type, s))
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return -1;
      }
 
@@ -465,7 +534,7 @@ struct_foreach_open (SLtype type, unsigned int num)
 
    (void) type;
 
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      return NULL;
 
    switch (num)
@@ -489,14 +558,14 @@ struct_foreach_open (SLtype type, unsigned int num)
 
    if (next_name == NULL)
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return NULL;
      }
 
    c = (SLang_Foreach_Context_Type *)SLmalloc (sizeof (SLang_Foreach_Context_Type));
    if (c == NULL)
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	SLang_free_slstring (next_name);
 	return NULL;
      }
@@ -514,7 +583,7 @@ static void struct_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
    if (c == NULL) return;
 
    SLang_free_slstring (c->next_field_name);
-   if (c->s != NULL) _pSLstruct_delete_struct (c->s);
+   if (c->s != NULL) SLang_free_struct (c->s);
    SLfree ((char *) c);
 }
 
@@ -531,7 +600,7 @@ static int struct_foreach (SLtype type, SLang_Foreach_Context_Type *c)
    if (c->s == NULL)
      return 0;			       /* done */
 
-   if (-1 == _pSLang_push_struct (c->s))
+   if (-1 == SLang_push_struct (c->s))
      return -1;
 
    /* Now get the next one ready for the next foreach loop */
@@ -553,7 +622,7 @@ static int struct_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 	  }
      }
 
-   _pSLstruct_delete_struct (c->s);
+   SLang_free_struct (c->s);
    c->s = next_s;
 
    /* keep going */
@@ -587,6 +656,14 @@ typedef struct
 }
 Unary_Op_Info_Type;
 
+typedef struct _Typecast_Info_Type
+{
+   SLang_Name_Type *typecast_fun;
+   SLtype totype;
+   struct _Typecast_Info_Type *next;
+}
+Typecast_Info_Type;
+  
 typedef struct _Struct_Info_Type
 {
    SLtype type;
@@ -596,6 +673,8 @@ typedef struct _Struct_Info_Type
    int unary_registered;
    Binary_Op_Info_Type *bi;
    Unary_Op_Info_Type *ui;
+   Typecast_Info_Type *ti;
+
    /* Other methods */
    SLang_Name_Type *destroy_method;
    SLang_Name_Type *string_method;
@@ -1130,7 +1209,7 @@ static void add_destroy_method (void)
 	return;
      }
    
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      {
 	SLang_free_function (f);
 	return;
@@ -1139,7 +1218,7 @@ static void add_destroy_method (void)
    if (s->destroy_method != NULL)
      SLang_free_function (s->destroy_method);
    s->destroy_method = SLang_copy_function (f);
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
 }
 
 static void add_string_method (SLtype *typep, SLang_Ref_Type *ref)
@@ -1160,6 +1239,95 @@ static void add_string_method (SLtype *typep, SLang_Ref_Type *ref)
    si->string_method = SLang_copy_function (f);
 }
 
+static Typecast_Info_Type *find_typecast (Struct_Info_Type *si, SLtype to)
+{
+   Typecast_Info_Type *ti = si->ti;
+
+   while (ti != NULL) 
+     {
+	if (ti->totype == to)
+	  return ti;
+	ti = ti->next;
+     }
+   return ti;
+}
+
+static int typecast_method (SLtype a_type, VOID_STAR ap, SLuindex_Type na,
+			    SLtype b_type, VOID_STAR bp)
+{
+   Struct_Info_Type *si;
+   Typecast_Info_Type *ti;
+   SLuindex_Type i;
+   SLang_Class_Type *acl, *bcl;
+   int (*apush) (SLtype, VOID_STAR);
+   int (*bpop) (SLtype, VOID_STAR);
+   unsigned int ainc, binc;
+   SLang_Name_Type *f;
+
+   if (NULL == (si = find_struct_info (a_type, 1)))
+     return -1;
+   
+   if ((NULL == (ti = find_typecast (si, b_type)))
+       || (NULL == (f = ti->typecast_fun)))
+     {
+	SLang_verror (SL_TYPE_MISMATCH, "Typecast method not found");
+	return -1;
+     }
+   
+   acl = _pSLclass_get_class (a_type);
+   bcl = _pSLclass_get_class (b_type);
+   apush = acl->cl_apush;
+   bpop = bcl->cl_apop;
+   ainc = acl->cl_sizeof_type;
+   binc = bcl->cl_sizeof_type;
+
+   for (i = 0; i < na; i++)
+     {
+	if ((-1 == SLang_start_arg_list ())
+	    || (-1 == (*apush) (a_type, ap))
+	    || (-1 == SLang_end_arg_list ())
+	    || (-1 == SLexecute_function (f))
+	    || (-1 == (*bpop)(b_type, bp)))
+	  return -1;
+
+	ap = (VOID_STAR) ((char *)ap + ainc);
+	bp = (VOID_STAR) ((char *)bp + binc);
+     }
+   
+   return 1;
+}
+   
+static void add_typecast_method (SLtype *fromtype, SLtype *totype, SLang_Ref_Type *ref)
+{
+   Struct_Info_Type *si;
+   SLang_Name_Type *f;
+   SLtype to = *totype, from = *fromtype;
+   Typecast_Info_Type *ti;
+
+   if (NULL == (f = SLang_get_fun_from_ref (ref)))
+     return;
+
+   if (NULL == (si = find_struct_info (from, 1)))
+     return;
+   
+   if (NULL != (ti = find_typecast (si, to)))
+     {
+	if (ti->typecast_fun != NULL)
+	  SLang_free_function (ti->typecast_fun);
+	ti->typecast_fun = SLang_copy_function (f);
+	return;
+     }
+
+   if (NULL == (ti = (Typecast_Info_Type *)SLmalloc (sizeof (Typecast_Info_Type))))
+     return;
+
+   ti->totype = to;
+   ti->typecast_fun = SLang_copy_function (f);
+   ti->next = si->ti;
+   si->ti = ti;
+   (void) SLclass_add_typecast (from, to, typecast_method, 1);
+}
+
 static int init_struct_with_user_methods (SLtype type, _pSLang_Struct_Type *s)
 {
    Struct_Info_Type *si;
@@ -1175,24 +1343,18 @@ static int init_struct_with_user_methods (SLtype type, _pSLang_Struct_Type *s)
 static int struct_sput (SLtype type, char *name)
 {
    _pSLang_Struct_Type *s;
-   _pSLstruct_Field_Type *f;
-   SLang_Object_Type obj;
 
    (void) type;
 
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      return -1;
 
-   if ((NULL == (f = pop_field (s, name, find_field)))
-       || (-1 == SLang_pop (&obj)))
+   if (-1 == pop_to_struct_field (s, name))
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return -1;
      }
-
-   SLang_free_object (&f->obj);
-   f->obj = obj;
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return 0;
 }
 
@@ -1204,17 +1366,17 @@ static int struct_sget (SLtype type, char *name)
 
    (void) type;
 
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      return -1;
 
    if (NULL == (f = pop_field (s, name, find_field)))
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return -1;
      }
 
    ret = _pSLpush_slang_obj (&f->obj);
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return ret;
 }
 
@@ -1322,7 +1484,7 @@ int _pSLstruct_define_typedef (void)
    if (-1 == SLang_pop_slstring (&type_name))
      return -1;
 
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      {
 	SLang_free_slstring (type_name);
 	return -1;
@@ -1331,16 +1493,16 @@ int _pSLstruct_define_typedef (void)
    if (NULL == (s1 = make_struct_shell (s, SLANG_STRUCT_TYPE)))
      {
 	SLang_free_slstring (type_name);
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return -1;
      }
 
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
 
    if (NULL == (cl = SLclass_allocate_class (type_name)))
      {
 	SLang_free_slstring (type_name);
-	_pSLstruct_delete_struct (s1);
+	SLang_free_struct (s1);
 	return -1;
      }
    SLang_free_slstring (type_name);
@@ -1502,7 +1664,7 @@ static void struct_set_field (void)
 	return;
      }
 
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      {
 	SLang_free_slstring (name);
 	SLang_free_object (&obj);
@@ -1511,7 +1673,7 @@ static void struct_set_field (void)
 
    if (NULL == (f = pop_field (s, name, find_field)))
      {
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	SLang_free_slstring (name);
 	SLang_free_object (&obj);
 	return;
@@ -1520,7 +1682,7 @@ static void struct_set_field (void)
    SLang_free_object (&f->obj);
    f->obj = obj;
 
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    SLang_free_slstring (name);
 }
 
@@ -1537,7 +1699,7 @@ static void set_struct_fields (void)
      return;
 
    n--;
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      {
 	SLdo_pop_n (n);
 	return;
@@ -1547,7 +1709,7 @@ static void set_struct_fields (void)
      {
 	SLdo_pop_n (n);
 	SLang_verror (SL_INVALID_PARM, "Too many values for structure");
-	_pSLstruct_delete_struct (s);
+	SLang_free_struct (s);
 	return;
      }
 
@@ -1566,7 +1728,7 @@ static void set_struct_fields (void)
 	n--;
      }
 
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
 }
 
 static void get_struct_field (char *name)
@@ -1627,6 +1789,7 @@ static SLang_Intrin_Fun_Type Struct_Table [] =
    MAKE_INTRINSIC_0("__add_binary", add_binary_op_intrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("__add_destroy", add_destroy_method, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_2("__add_string", add_string_method, SLANG_VOID_TYPE, SLANG_DATATYPE_TYPE, SLANG_REF_TYPE),
+   MAKE_INTRINSIC_3("__add_typecast", add_typecast_method, SLANG_VOID_TYPE, SLANG_DATATYPE_TYPE, SLANG_DATATYPE_TYPE, SLANG_REF_TYPE),
 
    /* MAKE_INTRINSIC_I("_create_struct", create_struct, SLANG_VOID_TYPE), */
    SLANG_END_INTRIN_FUN_TABLE
@@ -1700,7 +1863,7 @@ void _pSLstruct_pop_args (int *np)
 
 	s = data[i];
 	if (s != NULL)
-	  _pSLstruct_delete_struct (s);
+	  SLang_free_struct (s);
      }
 
    SLfree ((char *) data);
@@ -1785,7 +1948,7 @@ int SLang_pop_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
    if ((cfields == NULL) || (cs == NULL))
      return -1;
 
-   if (-1 == _pSLang_pop_struct (&s))
+   if (-1 == SLang_pop_struct (&s))
      return -1;
 
    cfield = cfields;
@@ -1793,20 +1956,30 @@ int SLang_pop_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
 
    while (NULL != (field_name = cfield->field_name))
      {
-	_pSLstruct_Field_Type *f;
-	SLang_Class_Type *cl;
-
-	if ((cfield->read_only == 0)
-	    && ((NULL == (f = pop_field (s, field_name, find_field_via_strcmp)))
-		|| (-1 == _pSLpush_slang_obj (&f->obj))
-		|| (NULL == (cl = _pSLclass_get_class (cfield->type)))
-		|| (-1 == (*cl->cl_apop)(cfield->type, (VOID_STAR) (cs_addr + cfield->offset)))))
-	  goto return_error;
+	if (cfield->read_only == 0)
+	  {
+	     _pSLstruct_Field_Type *f;
+	     SLang_Class_Type *cl;
+	     VOID_STAR addr = (VOID_STAR) (cs_addr + cfield->offset);
 
+	     if ((NULL == (f = pop_field (s, field_name, find_field_via_strcmp)))
+		 || (-1 == _pSLpush_slang_obj (&f->obj)))
+	       goto return_error;
+	
+	     if (cfield->type == SLANG_ARRAY_TYPE)
+	       {
+		  if (-1 == SLang_pop_array ((SLang_Array_Type **)addr, 1))
+		    goto return_error;
+	       }
+	     else if ((NULL == (cl = _pSLclass_get_class (cfield->type)))
+		      || (-1 == (*cl->cl_apop)(cfield->type, addr)))
+	       goto return_error;
+	  }
+	
 	cfield++;
      }
 
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return 0;
 
    return_error:
@@ -1815,7 +1988,7 @@ int SLang_pop_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
 	free_cstruct_field (cfield, cs);
 	cfield--;
      }
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return -1;
 }
 
@@ -1875,10 +2048,10 @@ int SLang_push_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
    if (NULL == (s = create_cstruct (cs, cfields)))
      return -1;
    
-   if (0 == _pSLang_push_struct (s))
+   if (0 == SLang_push_struct (s))
      return 0;
    
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return -1;
 }
 
@@ -1892,23 +2065,73 @@ int SLang_assign_cstruct_to_ref (SLang_Ref_Type *ref, VOID_STAR cs, SLang_CStruc
    if (0 == SLang_assign_to_ref (ref, SLANG_STRUCT_TYPE, (VOID_STAR) &s))
      return 0;
 
-   _pSLstruct_delete_struct (s);
+   SLang_free_struct (s);
    return -1;
 }
 
-/* Public functions */
-extern void SLang_free_struct (SLang_Struct_Type *s)
+/* Struct Field Reference */
+typedef struct
 {
-   _pSLstruct_delete_struct (s);
+   SLang_Struct_Type *s;
+   char *field_name;
+}
+Struct_Field_Ref_Type;
+
+static int struct_field_deref_assign (VOID_STAR vdata)
+{
+   Struct_Field_Ref_Type *data = (Struct_Field_Ref_Type *)vdata;
+   return pop_to_struct_field (data->s, data->field_name);
 }
 
-extern int SLang_push_struct (SLang_Struct_Type *s)
+static int struct_field_deref (VOID_STAR vdata)
 {
-   return _pSLang_push_struct (s);
+   Struct_Field_Ref_Type *frt = (Struct_Field_Ref_Type *)vdata;
+   _pSLstruct_Field_Type *f;
+
+   if (NULL == (f = pop_field (frt->s, frt->field_name, find_field)))
+     return -1;
+
+   return _pSLpush_slang_obj (&f->obj);
 }
 
-extern int SLang_pop_struct (SLang_Struct_Type **s)
+static void struct_field_ref_destroy (VOID_STAR vdata)
 {
-   return _pSLang_pop_struct (s);
+   Struct_Field_Ref_Type *frt = (Struct_Field_Ref_Type *)vdata;
+
+   SLang_free_slstring (frt->field_name);
+   SLang_free_struct (frt->s);
+}
+
+/* Stack: struct */
+int _pSLstruct_push_field_ref (char *name)
+{
+   SLang_Struct_Type *s;
+   Struct_Field_Ref_Type *frt;
+   SLang_Ref_Type *ref;
+   int ret;
+
+   if (-1 == SLang_pop_struct (&s))
+     return -1;
+
+   if (NULL == (name = SLang_create_slstring (name)))
+     {
+	SLang_free_struct (s);
+	return -1;
+     }
+   if (NULL == (ref = _pSLang_new_ref (sizeof (Struct_Field_Ref_Type))))
+     {
+	SLang_free_struct (s);
+	SLang_free_slstring (name);
+     }
+   frt = (Struct_Field_Ref_Type *) ref->data;
+   frt->s = s;
+   frt->field_name = name;
+   ref->deref = struct_field_deref;
+   ref->deref_assign = struct_field_deref_assign;
+   ref->destroy = struct_field_ref_destroy;
+
+   ret = SLang_push_ref (ref);
+   SLang_free_ref (ref);
+   return ret;
 }
 
