@@ -3,7 +3,7 @@
  */
 
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -232,27 +232,40 @@ static char *read_string_table (FILE *fp, SLterminfo_Type *t)
 
 static char *Terminfo_Dirs [] =
 {
-   NULL, /* $HOME/.terminfo */
-   NULL, /* $TERMINFO */
+   "", /* $HOME/.terminfo */
+   "", /* $TERMINFO */
+#ifdef MISC_TERMINFO_DIRS
+   MISC_TERMINFO_DIRS,
+#endif
    "/usr/share/terminfo",
    "/usr/lib/terminfo",
    "/usr/share/lib/terminfo",
    "/etc/terminfo",
+   "/lib/terminfo",
    "/usr/local/lib/terminfo",
-#ifdef MISC_TERMINFO_DIRS
-   MISC_TERMINFO_DIRS,
-#endif
-   ""
+   NULL,
 };
 
+void _pSLtt_tifreeent (SLterminfo_Type *t)
+{
+   if (t == NULL)
+     return;
+
+   SLfree ((char *)t->string_table);
+   SLfree ((char *)t->string_offsets);
+   SLfree ((char *)t->numbers);
+   SLfree ((char *)t->boolean_flags);
+   SLfree ((char *)t->terminal_names);
+   SLfree ((char *)t);
+}
+
 SLterminfo_Type *_pSLtt_tigetent (char *term)
 {
-   char *tidir;
-   int i;
+   char **tidirs, *tidir;
    FILE *fp = NULL;
    char file[1024];
    static char home_ti [1024];
-   char *home;
+   char *env;
    SLterminfo_Type *ti;
 
    if (
@@ -271,6 +284,7 @@ SLterminfo_Type *_pSLtt_tigetent (char *term)
      {
 	return NULL;
      }
+   memset ((char *)ti, 0, sizeof (SLterminfo_Type));
 
 #ifdef SLANG_UNTIC
    if (SLang_Untic_Terminfo_File != NULL)
@@ -283,68 +297,53 @@ SLterminfo_Type *_pSLtt_tigetent (char *term)
    /* If we are on a termcap based system, use termcap */
    if (0 == tcap_getent (term, ti)) return ti;
 
-   if (NULL != (home = _pSLsecure_getenv ("HOME")))
+   if (NULL != (env = _pSLsecure_getenv ("HOME")))
      {
-	strncpy (home_ti, home, sizeof (home_ti) - 11);
+	strncpy (home_ti, env, sizeof (home_ti) - 11);
 	home_ti [sizeof(home_ti) - 11] = 0;
 	strcat (home_ti, "/.terminfo");
 	Terminfo_Dirs [0] = home_ti;
      }
 
-   Terminfo_Dirs[1] = _pSLsecure_getenv ("TERMINFO");
-   i = 0;
-   while (1)
-     {
-	tidir = Terminfo_Dirs[i];
-	if (tidir != NULL)
-	  {
-	     if (*tidir == 0)
-	       break;		       /* last one */
+   if (NULL != (env = _pSLsecure_getenv ("TERMINFO")))
+     Terminfo_Dirs[1] = env;
 
-	     if (sizeof (file) > strlen (tidir) + 4 + strlen (term))
-	       {
-		  sprintf (file, "%s/%c/%s", tidir, *term, term);
-		  if (NULL != (fp = open_terminfo (file, ti)))
-		    break;
-	       }
+   tidirs = Terminfo_Dirs;
+   while (NULL != (tidir = *tidirs++))
+     {
+	if (*tidir == 0)
+	  continue;
+
+	if (sizeof (file) > strlen (tidir) + 4 + strlen (term))
+	  {
+	     sprintf (file, "%s/%c/%s", tidir, *term, term);
+	     if (NULL != (fp = open_terminfo (file, ti)))
+	       break;
 	  }
-	i++;
      }
 
 #ifdef SLANG_UNTIC
    fp_open_label:
 #endif
 
-   if (fp != NULL)
+   if (fp == NULL)
      {
-	if (NULL != read_terminal_names (fp, ti))
-	  {
-	     if (NULL != read_boolean_flags (fp, ti))
-	       {
-		  if (NULL != read_numbers (fp, ti))
-		    {
-		       if (NULL != read_string_offsets (fp, ti))
-			 {
-			    if (NULL != read_string_table (fp, ti))
-			      {
-				 /* success */
-				 fclose (fp);
-				 ti->flags = SLTERMINFO;
-				 return ti;
-			      }
-			    SLfree ((char *)ti->string_offsets);
-			 }
-		       SLfree ((char *)ti->numbers);
-		    }
-		  SLfree ((char *)ti->boolean_flags);
-	       }
-	     SLfree ((char *)ti->terminal_names);
-	  }
-	fclose (fp);
+	SLfree ((char *) ti);
+	return NULL;
      }
 
-   SLfree ((char *)ti);
-   return NULL;
+   ti->flags = SLTERMINFO;
+   if ((NULL == read_terminal_names (fp, ti))
+       || (NULL == read_boolean_flags (fp, ti))
+       || (NULL == read_numbers (fp, ti))
+       || (NULL == read_string_offsets (fp, ti))
+       || (NULL == read_string_table (fp, ti)))
+     {
+	_pSLtt_tifreeent (ti);
+	ti = NULL;
+     }
+   (void) fclose (fp);
+   return ti;
 }
 
 #ifdef SLANG_UNTIC
