@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -123,6 +123,9 @@ static char *map_token_to_string (_pSLang_Token_Type *tok)
       case SEMICOLON_TOKEN: s = ";"; break;
       case COLON_TOKEN: s = ":"; break;
 
+      case ARRAY_TOKEN: s = "["; break;
+      case DOT_TOKEN: s = "."; break;
+
 #if SLANG_HAS_FLOAT
       case FLOAT_TOKEN:
       case DOUBLE_TOKEN:
@@ -203,7 +206,7 @@ static SLCONST char Operators [31][4] =
      {'#',	0,	0, POUND_TOKEN},
 #define OFS_BAND	2
      {'&',	0,	0, BAND_TOKEN},
-     {'&',	'&',	0, EOF_TOKEN},
+     {'&',	'&',	0, SC_AND_TOKEN},
      {'&',	'=',	0, BANDEQS_TOKEN},
 #define OFS_STAR	5
      {'*',	0,	0, TIMES_TOKEN},
@@ -237,7 +240,7 @@ static SLCONST char Operators [31][4] =
      {'^',	0,	0, POW_TOKEN},
 #define OFS_BOR		26
      {'|',	0,	0, BOR_TOKEN},
-     {'|',	'|',	0, EOF_TOKEN},
+     {'|',	'|',	0, SC_OR_TOKEN},
      {'|',	'=',	0, BOREQS_TOKEN},
 #define OFS_BNOT	29
      {'~',	0,	0, BNOT_TOKEN},
@@ -763,7 +766,7 @@ static int get_string_token (_pSLang_Token_Type *tok, unsigned char quote_char,
 
 	if (len == (MAX_TOKEN_LEN - 1))
 	  {
-	     _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "String too long for buffer", NULL, 0);
+	     _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "Literal string exceeds the maximum allowable size--- use concatenation", NULL, 0);
 	     return (tok->type = EOF_TOKEN);
 	  }
 
@@ -1277,7 +1280,7 @@ int SLns_load_string (char *string, char *ns_name)
 
    data.ptr = data.string = string;
    if ((-1 == (ret = SLang_load_object (x)))
-       && (SLang_Traceback))
+       && (SLang_Traceback & SL_TB_FULL))
      _pSLerr_traceback_msg ("Traceback: called from eval: %s\n", string);
 
    SLang_free_slstring (string);
@@ -1344,7 +1347,7 @@ int SLns_load_file (char *f, char *ns_name)
    if (f == NULL) 
      name = SLang_create_slstring ("<stdin>");
    else
-     name = _pSLpath_find_file (f);
+     name = _pSLpath_find_file (f, 1);
 	
    if (name == NULL)
      return -1;
@@ -1527,6 +1530,7 @@ void _pSLcompile_byte_compiled (void)
 	   case _STRUCT_MINUSMINUS_TOKEN:
 	   case _STRUCT_POST_PLUSPLUS_TOKEN:
 	   case _STRUCT_PLUSPLUS_TOKEN:
+	   case _STRUCT_FIELD_REF_TOKEN:
 	     if (NULL == (ebuf = check_byte_compiled_token (buf)))
 	       return;
 	     tok.v.s_val = buf;
@@ -1743,6 +1747,7 @@ static void byte_compile_token (_pSLang_Token_Type *tok)
       case _STRUCT_MINUSMINUS_TOKEN:
       case _STRUCT_POST_PLUSPLUS_TOKEN:
       case _STRUCT_PLUSPLUS_TOKEN:
+      case _STRUCT_FIELD_REF_TOKEN:
 	strcpy (b3, tok->v.s_val);
 	break;
 
