@@ -1,6 +1,6 @@
 /* Basic type operations for S-Lang */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -359,26 +359,58 @@ static int string_acopy (SLtype unused, VOID_STAR src_sptr, VOID_STAR dest_sptr)
 struct _pSLang_Foreach_Context_Type
 {
    char *string;
-   unsigned int n;
+   unsigned char *s, *smax;	       /* pointers into string */
+   int using_chars;
 };
 
 static SLang_Foreach_Context_Type *
 string_foreach_open (SLtype type, unsigned int num)
 {
    char *s;
+   char *u;
    SLang_Foreach_Context_Type *c;
+   int using_chars = 0;
 
    (void) type;
-   if (num != 0)
-     {
-	SLang_verror (SL_NOT_IMPLEMENTED,
-		      "'foreach using' form not supported by String_Type");
-	SLdo_pop_n (num + 1);
-	return NULL;
-     }
+
    if (-1 == SLang_pop_slstring (&s))
      return NULL;
 
+   switch (num)
+     {
+      case 1:
+	if (-1 == SLang_pop_slstring (&u))
+	  {
+	     SLang_free_slstring (s);
+	     return NULL;
+	  }
+	if (0 == strcmp (u, "chars"))
+	  using_chars = 1;
+	else if (0 == strcmp (u, "bytes"))
+	  using_chars = 0;
+	else
+	  {
+	     SLang_verror (SL_InvalidParm_Error, "Expected foreach (String_Type) using (chars|bytes)");
+	     SLang_free_slstring (u);
+	     SLang_free_slstring (s);
+	     return NULL;
+	  }
+	SLang_free_slstring (u);
+   	break;
+
+      case 0:
+	using_chars = 0;
+	break;
+      default:
+	SLang_verror (SL_NumArgs_Error,
+		      "'foreach (String_Type) using' requires single control value (chars|bytes)");
+	return NULL;
+     }
+   
+   /* In UTF-8 mode, chars and bytes are synonymous */
+   if (_pSLinterp_UTF8_Mode == 0)
+     using_chars = 0;
+
    c = (SLang_Foreach_Context_Type *)SLmalloc (sizeof (SLang_Foreach_Context_Type));
    if (c == NULL)
      {
@@ -388,7 +420,9 @@ string_foreach_open (SLtype type, unsigned int num)
 
    memset ((char *) c, 0, sizeof (SLang_Foreach_Context_Type));
    c->string = s;
-
+   c->s = (unsigned char *) s;
+   c->smax = (unsigned char *)s + strlen (s);
+   c->using_chars = using_chars;
    return c;
 }
 
@@ -402,18 +436,42 @@ static void string_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
 
 static int string_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 {
-   char ch;
-
+   unsigned char ch;
+   SLwchar_Type wch;
+   unsigned char *s, *s1, *smax;
+   
    (void) type;
-   ch = c->string[c->n];
-   if (ch == 0)
-     return 0;			       /* done */
 
-   c->n += 1;
+   s = c->s;
+   smax = c->smax;
+   if (s == smax)
+     return 0;
 
-   if (-1 == SLclass_push_int_obj (SLANG_INT_TYPE, ch))
+   if (c->using_chars == 0)
+     {
+	ch = (unsigned char) *s++;
+	c->s = s;
+
+	if (-1 == SLclass_push_char_obj (SLANG_UCHAR_TYPE, ch))
+	  return -1;
+
+	return 1;
+     }
+   s1 = SLutf8_decode (s, smax, &wch, NULL);
+   if (s1 == NULL)
+     {
+	int iwch = (int) *s;
+	c->s = s + 1;
+	/* Invalid encoded char-- return it as a negative int */
+	if (-1 == SLang_push_int (-iwch))
+	  return -1;
+	
+	return 1;
+     }
+   c->s = s1;
+   if (-1 == SLang_push_wchar (wch))
      return -1;
-
+   
    return 1;
 }
 
@@ -587,75 +645,62 @@ _pSLString_List_Type *_pSLstring_list_new (unsigned int max_num, unsigned int de
 
 /* Ref type */
 
+int _pSLang_deref_assign (SLang_Ref_Type *ref)
+{
+   return ref->deref_assign (ref->data);
+}
+
 int SLang_pop_ref (SLang_Ref_Type **ref)
 {
    return SLclass_pop_ptr_obj (SLANG_REF_TYPE, (VOID_STAR *)ref);
 }
- 
-static SLang_Ref_Type *create_ref_to_nametype (SLang_Name_Type *nt, int is_global)
+
+int SLang_push_ref (SLang_Ref_Type *ref)
 {
-   SLang_Ref_Type *r;
-
-   r = (SLang_Ref_Type *) SLmalloc (sizeof (SLang_Ref_Type));
-   if (r == NULL) return NULL;
-
-   r->is_global = is_global;
-   r->v.nt = nt;
-   return r;
+   ref->num_refs++;
+   if (0 == SLclass_push_ptr_obj (SLANG_REF_TYPE, (VOID_STAR) ref))
+     return 0;
+   ref->num_refs--;
+   return -1;
 }
 
-int SLang_assign_nametype_to_ref (SLang_Ref_Type *ref, SLang_Name_Type *nt)
+void SLang_free_ref (SLang_Ref_Type *ref)
 {
-   SLang_Ref_Type *r;
-
-   if ((nt == NULL) || (ref == NULL))
-     return -1;
+   if (ref == NULL)
+     return;
    
-   if (NULL == (r = create_ref_to_nametype (nt, 1)))
-     return -1;
-
-   if (-1 == SLang_assign_to_ref (ref, SLANG_REF_TYPE, (VOID_STAR) &r))
+   if (ref->num_refs > 1)
      {
-	SLang_free_ref (r);
-	return -1;
+	ref->num_refs--;
+	return;
      }
-   SLang_free_ref (r);
-   return 0;
+
+   if (ref->destroy != NULL)
+     (*ref->destroy)(ref->data);
+   SLfree ((char *)ref->data);
+   SLfree ((char *)ref);
 }
 
-/* Note: This is ok if nt is NULL.  Some routines rely on this behavior */
-static int push_nametype_as_ref (SLang_Name_Type *nt, int is_global)
+SLang_Ref_Type *_pSLang_new_ref (unsigned int sizeof_data)
 {
-   SLang_Ref_Type *r;
-
-   if (nt == NULL)
-     return SLang_push_null ();
-
-   r = create_ref_to_nametype (nt, is_global);
-   if (r == NULL) return -1;
-
-   if (-1 == SLclass_push_ptr_obj (SLANG_REF_TYPE, (VOID_STAR) r))
+   SLang_Ref_Type *ref;
+   
+   if (NULL == (ref = (SLang_Ref_Type *)SLcalloc (1, sizeof (SLang_Ref_Type))))
+     return NULL;
+   if (NULL == (ref->data = (VOID_STAR)SLcalloc (1, sizeof_data)))
      {
-	SLfree ((char *) r);
-	return -1;
+	SLfree ((char *)ref);
+	return NULL;
      }
-   return 0;
-}
-
-int _pSLang_push_ref (int is_global, VOID_STAR ptr)
-{
-   return push_nametype_as_ref ((SLang_Name_Type *)ptr, is_global);
+   ref->num_refs = 1;
+   ref->sizeof_data = sizeof_data;
+   return ref;
 }
 
 static void ref_destroy (SLtype type, VOID_STAR ptr)
 {
    (void) type;
-   SLfree ((char *) *(SLang_Ref_Type **)ptr);
-}
-
-void SLang_free_ref (SLang_Ref_Type *ref)
-{
-   SLfree ((char *) ref);
+   SLang_free_ref (*(SLang_Ref_Type **)ptr);
 }
 
 static int ref_push (SLtype type, VOID_STAR ptr)
@@ -669,7 +714,7 @@ static int ref_push (SLtype type, VOID_STAR ptr)
    if (ref == NULL)
      return SLang_push_null ();
 
-   return _pSLang_push_ref (ref->is_global, (VOID_STAR) ref->v.nt);
+   return SLang_push_ref (ref);
 }
 
 int SLang_assign_to_ref (SLang_Ref_Type *ref, SLtype type, VOID_STAR v)
@@ -715,22 +760,9 @@ static char *ref_string (SLtype type, VOID_STAR ptr)
 
    (void) type;
    ref = *(SLang_Ref_Type **) ptr;
-   if (ref->is_global)
-     {
-	char *name, *s;
-
-	name = ref->v.nt->name;
-	if ((name != NULL)
-	    && (NULL != (s = SLmalloc (strlen(name) + 2))))
-	  {
-	     *s = '&';
-	     strcpy (s + 1, name);
-	     return s;
-	  }
-	
-	return NULL;
-     }
-   return SLmake_string ("Local Variable Reference");
+   if (ref->string != NULL)
+     return ref->string (ref->data);
+   return SLmake_string ("Ref_Type");
 }
 
 static int ref_dereference (SLtype unused, VOID_STAR ptr)
@@ -744,10 +776,10 @@ static int ref_cmp (SLtype type, VOID_STAR a, VOID_STAR b, int *c)
    SLang_Ref_Type *ra, *rb;
 
    (void) type;
-   
+
    ra = *(SLang_Ref_Type **)a;
    rb = *(SLang_Ref_Type **)b;
-   
+
    if (ra == NULL)
      {
 	if (rb == NULL) *c = 0;
@@ -759,16 +791,19 @@ static int ref_cmp (SLtype type, VOID_STAR a, VOID_STAR b, int *c)
 	*c = 1;
 	return 0;
      }
-	
-   if (ra->v.nt == rb->v.nt)
-     *c = 0;
-   else *c = strcmp (ra->v.nt->name, rb->v.nt->name);
+   if (ra->sizeof_data != rb->sizeof_data)
+     {
+	*c = (int)ra->sizeof_data - (int)rb->sizeof_data;
+	return 0;
+     }
+
+   *c = memcmp (ra->data, rb->data, ra->sizeof_data);
    return 0;
 }
 
 int SLang_push_function (SLang_Name_Type *nt)
 {
-   return push_nametype_as_ref (nt, 1);
+   return _pSLang_push_nt_as_ref (nt);
 }
    
 SLang_Name_Type *SLang_pop_function (void)
