@@ -1,6 +1,6 @@
 /* slwclut.c: wide character lookup tables */
 /*
-Copyright (C) 2004, 2005, 2006 John E. Davis
+Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -348,7 +348,7 @@ static int is_posix_charclass (SLuchar_Type **up, SLuchar_Type *umax, SLwchar_Ty
 
    u1 = u;
    while ((u1 < umax)
-	  && (*u1 > 'a')
+	  && (*u1 >= 'a')
 	  && (*u1 <= 'z'))
      u1++;
 
